"""
This script contains the functions/classes to compute loss and metrics used to train and evaluate the performance of the model.
"""

import numpy as np
import torch
from skimage.metrics import structural_similarity
from torchmetrics.image import MultiScaleStructuralSimilarityIndexMeasure

from careamics.models.lvae.utils import allow_numpy


class RunningPSNR:
    """
    This class allows to compute the running PSNR during validation step in training.
    In this way it is possible to compute the PSNR on the entire validation set one batch at the time.
    """

    def __init__(self):
        # number of elements seen so far during the epoch
        self.N = None
        # running sum of the MSE over the self.N elements seen so far
        self.mse_sum = None
        # running max and min values of the self.N target images seen so far
        self.max = self.min = None
        self.reset()

    def reset(self):
        """
        Used to reset the running PSNR (usually called at the end of each epoch).
        """
        self.mse_sum = 0
        self.N = 0
        self.max = self.min = None

    def update(self, rec: torch.Tensor, tar: torch.Tensor) -> None:
        """
        Given a batch of reconstructed and target images, it updates the MSE and.

        Parameters
        ----------
        rec: torch.Tensor
            Batch of reconstructed images (B, H, W).
        tar: torch.Tensor
            Batch of target images (B, H, W).
        """
        ins_max = torch.max(tar).item()
        ins_min = torch.min(tar).item()
        if self.max is None:
            assert self.min is None
            self.max = ins_max
            self.min = ins_min
        else:
            self.max = max(self.max, ins_max)
            self.min = min(self.min, ins_min)

        mse = (rec - tar) ** 2
        elementwise_mse = torch.mean(mse.view(len(mse), -1), dim=1)
        self.mse_sum += torch.nansum(elementwise_mse)
        self.N += len(elementwise_mse) - torch.sum(torch.isnan(elementwise_mse))

    def get(self):
        """
        The get the actual PSNR value given the running statistics.
        """
        if self.N == 0 or self.N is None:
            return None
        rmse = torch.sqrt(self.mse_sum / self.N)
        return 20 * torch.log10((self.max - self.min) / rmse)


def zero_mean(x):
    return x - torch.mean(x, dim=1, keepdim=True)


def fix_range(gt, x):
    a = torch.sum(gt * x, dim=1, keepdim=True) / (torch.sum(x * x, dim=1, keepdim=True))
    return x * a


def fix(gt, x):
    gt_ = zero_mean(gt)
    return fix_range(gt_, zero_mean(x))


def _PSNR_internal(gt, pred, range_=None):
    if range_ is None:
        range_ = torch.max(gt, dim=1).values - torch.min(gt, dim=1).values

    mse = torch.mean((gt - pred) ** 2, dim=1)
    return 20 * torch.log10(range_ / torch.sqrt(mse))


@allow_numpy
def PSNR(gt, pred, range_=None):
    """
    Compute PSNR.

    Parameters
    ----------
    gt: array
        Ground truth image.
    pred: array
        Predicted image.
    """
    assert len(gt.shape) == 3, "Images must be in shape: (batch,H,W)"

    gt = gt.view(len(gt), -1)
    pred = pred.view(len(gt), -1)
    return _PSNR_internal(gt, pred, range_=range_)


@allow_numpy
def RangeInvariantPsnr(gt: torch.Tensor, pred: torch.Tensor):
    """
    NOTE: Works only for grayscale images.
    Adapted from https://github.com/juglab/ScaleInvPSNR/blob/master/psnr.py
    It rescales the prediction to ensure that the prediction has the same range as the ground truth.
    """
    assert len(gt.shape) == 3, "Images must be in shape: (batch,H,W)"
    gt = gt.view(len(gt), -1)
    pred = pred.view(len(gt), -1)
    ra = (torch.max(gt, dim=1).values - torch.min(gt, dim=1).values) / torch.std(
        gt, dim=1
    )
    gt_ = zero_mean(gt) / torch.std(gt, dim=1, keepdim=True)
    return _PSNR_internal(zero_mean(gt_), fix(gt_, pred), ra)


def _avg_psnr(target, prediction, psnr_fn):
    output = np.mean(
        [
            psnr_fn(target[i : i + 1], prediction[i : i + 1]).item()
            for i in range(len(prediction))
        ]
    )
    return round(output, 2)


def avg_range_inv_psnr(target, prediction):
    return _avg_psnr(target, prediction, RangeInvariantPsnr)


def avg_psnr(target, prediction):
    return _avg_psnr(target, prediction, PSNR)


def compute_masked_psnr(mask, tar1, tar2, pred1, pred2):
    mask = mask.astype(bool)
    mask = mask[..., 0]
    tmp_tar1 = tar1[mask].reshape((len(tar1), -1, 1))
    tmp_pred1 = pred1[mask].reshape((len(tar1), -1, 1))
    tmp_tar2 = tar2[mask].reshape((len(tar2), -1, 1))
    tmp_pred2 = pred2[mask].reshape((len(tar2), -1, 1))
    psnr1 = avg_range_inv_psnr(tmp_tar1, tmp_pred1)
    psnr2 = avg_range_inv_psnr(tmp_tar2, tmp_pred2)
    return psnr1, psnr2


def avg_ssim(target, prediction):
    ssim = [
        structural_similarity(
            target[i], prediction[i], data_range=(target[i].max() - target[i].min())
        )
        for i in range(len(target))
    ]
    return np.mean(ssim), np.std(ssim)


@allow_numpy
def range_invariant_multiscale_ssim(gt_, pred_):
    """
    Computes range invariant multiscale ssim for one channel.
    This has the benefit that it is invariant to scalar multiplications in the prediction.
    """
    shape = gt_.shape
    gt_ = torch.Tensor(gt_.reshape((shape[0], -1)))
    pred_ = torch.Tensor(pred_.reshape((shape[0], -1)))
    gt_ = zero_mean(gt_)
    pred_ = zero_mean(pred_)
    pred_ = fix(gt_, pred_)
    pred_ = pred_.reshape(shape)
    gt_ = gt_.reshape(shape)

    ms_ssim = MultiScaleStructuralSimilarityIndexMeasure(
        data_range=gt_.max() - gt_.min()
    )
    return ms_ssim(torch.Tensor(pred_[:, None]), torch.Tensor(gt_[:, None])).item()


def compute_multiscale_ssim(gt_, pred_, range_invariant=True):
    """
    Computes multiscale ssim for each channel.
    Args:
    gt_: ground truth image with shape (N, H, W, C)
    pred_: predicted image with shape (N, H, W, C)
    range_invariant: whether to use range invariant multiscale ssim
    """
    ms_ssim_values = {i: None for i in range(gt_.shape[-1])}
    for ch_idx in range(gt_.shape[-1]):
        tar_tmp = gt_[..., ch_idx]
        pred_tmp = pred_[..., ch_idx]
        if range_invariant:
            ms_ssim_values[ch_idx] = range_invariant_multiscale_ssim(tar_tmp, pred_tmp)
        else:
            ms_ssim = MultiScaleStructuralSimilarityIndexMeasure(
                data_range=tar_tmp.max() - tar_tmp.min()
            )
            ms_ssim_values[ch_idx] = ms_ssim(
                torch.Tensor(pred_tmp[:, None]), torch.Tensor(tar_tmp[:, None])
            ).item()

    output = [ms_ssim_values[i] for i in range(gt_.shape[-1])]
    return output



"""Losses supported by CAREamics."""

from careamics.utils import BaseEnum


# TODO register loss with custom_loss decorator?
class SupportedLoss(str, BaseEnum):
    """Supported losses.

    Attributes
    ----------
    MSE : str
        Mean Squared Error loss.
    MAE : str
        Mean Absolute Error loss.
    N2V : str
        Noise2Void loss.
    """

    MSE = "mse"
    MAE = "mae"
    N2V = "n2v"
    # PN2V = "pn2v"
    # HDN = "hdn"
    MUSPLIT = "musplit"
    DENOISPLIT = "denoisplit"
    DENOISPLIT_MUSPLIT = "denoisplit_musplit"
    # CE = "ce"
    # DICE = "dice"



"""Version utility."""

import logging

from careamics import __version__

logger = logging.getLogger(__name__)


def get_careamics_version() -> str:
    """Get clean CAREamics version.

    This method returns the latest `Major.Minor.Patch` version of CAREamics, removing
    any local version identifier.

    Returns
    -------
    str
        Clean CAREamics version.
    """
    parts = __version__.split(".")

    # for local installs that do not detect the latest versions via tags
    # (typically our CI will install `0.1.devX<hash>` versions)
    if "dev" in parts[-1]:
        parts[-1] = "*"
        clean_version = ".".join(parts[:3])

        logger.warning(
            f"Your CAREamics version seems to be a locally modified version "
            f"({__version__}). The recorded version for loading models will be "
            f"{clean_version}, which may not exist. If you want to ensure "
            f"exporting the model with an existing version, please install the "
            f"closest CAREamics version from PyPI or conda-forge."
        )

    # Remove any local version identifier)
    return ".".join(parts[:3])



import numpy as np
import pytest

from careamics.dataset.patching.sequential_patching import (
    _compute_number_of_patches,
    _compute_overlap,
    _compute_patch_steps,
    _compute_patch_views,
    extract_patches_sequential,
)


def check_extract_patches_sequential(array: np.ndarray, patch_size: tuple):
    """Check that the patches are extracted correctly.

    The array should have been generated using np.arange and np.reshape."""
    patches, _ = extract_patches_sequential(array, patch_size=patch_size)

    # check patch shape
    assert patches.shape[2:] == patch_size

    # check that all values are covered by the patches
    n_max = np.prod(array.shape)  # maximum value in the array
    unique = np.unique(np.array(patches))  # unique values in the patches
    assert len(unique) == n_max


@pytest.mark.parametrize(
    "patch_size",
    [
        (2, 2),
        (4, 2),
        (4, 8),
        (8, 8),
    ],
)
def test_extract_patches_sequential_2d(array_2D, patch_size):
    """Test extracting patches sequentially in 2D."""
    check_extract_patches_sequential(array_2D, patch_size)


@pytest.mark.parametrize(
    "patch_size",
    [
        (2, 2, 4),
        (4, 2, 2),
        (2, 8, 4),
        (4, 8, 8),
    ],
)
def test_extract_patches_sequential_3d(array_3D, patch_size):
    """Test extracting patches sequentially in 3D.

    The 3D array is a fixture of shape (1, 8, 16, 16)."""
    # TODO changed the fixture to (1, 8, 16, 16), uneven shape doesnt work. We need to
    # discuss the function or the test cases
    check_extract_patches_sequential(array_3D, patch_size)


@pytest.mark.parametrize(
    "patch_size",
    [
        (4, 4),
        (8, 8),
    ],
)
def test_extract_patches_sequential_2d_supervised(array_2D, patch_size):
    """Test extracting patches sequentially in 2D with target."""
    patches, targets = extract_patches_sequential(
        array_2D, patch_size=patch_size, target=array_2D
    )

    # Check that the same region is extracted in the patches and targets
    assert np.array_equal(patches, targets)


@pytest.mark.parametrize(
    "shape, patch_sizes, expected",
    [
        ((1, 3, 10, 10), (1, 3, 10, 5), (1, 1, 1, 2)),
        ((1, 1, 9, 9), (1, 1, 4, 3), (1, 1, 3, 3)),
        ((1, 3, 10, 9), (1, 3, 3, 5), (1, 1, 4, 2)),
        ((1, 1, 5, 9, 10), (1, 1, 2, 3, 5), (1, 1, 3, 3, 2)),
    ],
)
def test_compute_number_of_patches(shape, patch_sizes, expected):
    """Test computing number of patches"""
    assert _compute_number_of_patches(shape, patch_sizes) == expected


@pytest.mark.parametrize(
    "shape, patch_sizes, expected",
    [
        ((1, 3, 10, 10), (1, 3, 10, 5), (0, 0, 0, 0)),
        ((1, 1, 9, 9), (1, 1, 4, 3), (0, 0, 2, 0)),
        ((1, 3, 10, 10, 9), (1, 3, 2, 3, 5), (0, 0, 0, 1, 1)),
    ],
)
def test_compute_overlap(shape, patch_sizes, expected):
    """Test computing overlap between patches"""
    assert _compute_overlap(shape, patch_sizes) == expected


@pytest.mark.parametrize("dims", [2, 3])
@pytest.mark.parametrize("patch_size", [2, 3])
@pytest.mark.parametrize("overlap", [0, 1, 4])
def test_compute_patch_steps(dims, patch_size, overlap):
    """Test computing patch steps"""
    patch_sizes = (patch_size,) * dims
    overlaps = (overlap,) * dims
    expected = (min(patch_size - overlap, patch_size),) * dims

    assert _compute_patch_steps(patch_sizes, overlaps) == expected


def check_compute_reshaped_view(array: np.ndarray, window_shape, steps):
    """Check the number of patches"""

    output_shape = (-1, *window_shape)

    # compute views
    output = _compute_patch_views(array, window_shape, steps, output_shape)

    # check the number of patches
    n_patches = [
        np.ceil((array.shape[i] - window_shape[i] + 1) / steps[i]).astype(int)
        for i in range(len(window_shape))
    ]
    assert output.shape == (np.prod(n_patches), *window_shape)


@pytest.mark.parametrize(
    "window_shape, steps",
    [
        ((1, 3, 5, 5), (1, 1, 1, 1)),
        ((1, 3, 5, 5), (1, 1, 2, 3)),
        ((1, 3, 5, 7), (1, 1, 1, 1)),
    ],
)
def test_compute_reshaped_view_2d(array_2D, window_shape, steps):
    """Test computing reshaped view of an array of shape (1, 10, 9)."""
    check_compute_reshaped_view(array_2D, window_shape, steps)


@pytest.mark.parametrize(
    "window_shape, steps",
    [
        ((1, 3, 1, 5, 5), (1, 1, 2, 1, 2)),
        ((1, 3, 2, 5, 5), (1, 1, 2, 3, 4)),
        ((1, 3, 3, 7, 8), (1, 1, 1, 1, 3)),
    ],
)
def test_compute_reshaped_view_3d(array_3D, window_shape, steps):
    """Test computing reshaped view of an array of shape (1, 5, 10, 9)."""
    check_compute_reshaped_view(array_3D, window_shape, steps)



from collections.abc import Sequence

from numpy.typing import NDArray

from .image_stack import ImageStack


class PatchExtractor:
    """
    A class for extracting patches from multiple image stacks.
    """

    def __init__(self, image_stacks: Sequence[ImageStack]):
        self.image_stacks: list[ImageStack] = list(image_stacks)

    def extract_patch(
        self,
        data_idx: int,
        sample_idx: int,
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        return self.image_stacks[data_idx].extract_patch(
            sample_idx=sample_idx, coords=coords, patch_size=patch_size
        )

    @property
    def shape(self):
        return [stack.data_shape for stack in self.image_stacks]



"""Module to get write functions."""

from pathlib import Path
from typing import Literal, Protocol

from numpy.typing import NDArray

from careamics.config.support import SupportedData

from .tiff import write_tiff

SupportedWriteType = Literal["tiff", "custom"]


# This is very strict, arguments have to be called file_path & img
# Alternative? - doesn't capture *args & **kwargs
# WriteFunc = Callable[[Path, NDArray], None]
class WriteFunc(Protocol):
    """Protocol for type hinting write functions."""

    def __call__(self, file_path: Path, img: NDArray, *args, **kwargs) -> None:
        """
        Type hinted callables must match this function signature (not including self).

        Parameters
        ----------
        file_path : pathlib.Path
            Path to file.
        img : numpy.ndarray
            Image data to save.
        *args
            Other positional arguments.
        **kwargs
            Other keyword arguments.
        """


WRITE_FUNCS: dict[SupportedData, WriteFunc] = {
    SupportedData.TIFF: write_tiff,
}


def get_write_func(data_type: SupportedWriteType) -> WriteFunc:
    """
    Get the write function for the data type.

    Parameters
    ----------
    data_type : {"tiff", "custom"}
        Data type.

    Returns
    -------
    callable
        Write function.
    """
    # error raised here if not supported
    data_type_ = SupportedData(data_type)  # new variable for mypy
    # error if no write func.
    if data_type_ not in WRITE_FUNCS:
        raise NotImplementedError(f"No write function for data type '{data_type}'.")

    return WRITE_FUNCS[data_type_]



"""Plotting utilities."""

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import torch
from numpy.typing import NDArray

from careamics.models.lvae.noise_models import GaussianMixtureNoiseModel


def plot_noise_model_probability_distribution(
    noise_model: GaussianMixtureNoiseModel,
    signalBinIndex: int,
    histogram: NDArray,
    channel: Optional[str] = None,
    number_of_bins: int = 100,
) -> None:
    """Plot probability distribution P(x|s) for a certain ground truth signal.

    Predictions from both Histogram and GMM-based
    Noise models are displayed for comparison.

    Parameters
    ----------
    noise_model : GaussianMixtureNoiseModel
        Trained GaussianMixtureNoiseModel.
    signalBinIndex : int
        Index of signal bin. Values go from 0 to number of bins (`n_bin`).
    histogram : NDArray
        Histogram based noise model.
    channel : Optional[str], optional
        Channel name used for plotting. Default is None.
    number_of_bins : int, optional
        Number of bins in the resulting histogram. Default is 100.
    """
    min_signal = noise_model.min_signal.item()
    max_signal = noise_model.max_signal.item()
    bin_size = (max_signal - min_signal) / number_of_bins

    query_signal_normalized = signalBinIndex / number_of_bins
    query_signal = query_signal_normalized * (max_signal - min_signal) + min_signal
    query_signal += bin_size / 2
    query_signal = torch.tensor(query_signal)

    query_observations = torch.arange(min_signal, max_signal, bin_size)
    query_observations += bin_size / 2

    likelihoods = noise_model.likelihood(
        observations=query_observations, signals=query_signal
    ).numpy()

    plt.figure(figsize=(12, 5))
    if channel:
        plt.suptitle(f"Noise model for channel {channel}")
    else:
        plt.suptitle("Noise model")

    plt.subplot(1, 2, 1)
    plt.xlabel("Observation Bin")
    plt.ylabel("Signal Bin")
    plt.imshow(histogram**0.25, cmap="gray")
    plt.axhline(y=signalBinIndex + 0.5, linewidth=5, color="blue", alpha=0.5)

    plt.subplot(1, 2, 2)
    plt.plot(
        query_observations,
        likelihoods,
        label="GMM : " + " signal = " + str(np.round(query_signal, 2)),
        marker=".",
        color="red",
        linewidth=2,
    )
    plt.xlabel("Observations (x) for signal s = " + str(query_signal))
    plt.ylabel("Probability Density")
    plt.title("Probability Distribution P(x|s) at signal =" + str(query_signal))
    plt.legend()



"""Test `WriteImage` class."""

from pathlib import Path
from unittest.mock import DEFAULT, Mock, patch

import numpy as np
import pytest
from pytorch_lightning import LightningModule, Trainer
from torch.utils.data import DataLoader

from careamics.dataset import IterablePredDataset
from careamics.file_io import WriteFunc
from careamics.lightning.callbacks.prediction_writer_callback.write_strategy import (
    WriteImage,
)


@pytest.fixture
def write_func():
    """Mock `WriteFunc`."""
    return Mock(spec=WriteFunc)


@pytest.fixture
def write_image_strategy(write_func) -> WriteImage:
    """
    Initialized `CacheTiles` class.

    Parameters
    ----------
    write_func : WriteFunc
        Write function. (Comes from fixture).

    Returns
    -------
    CacheTiles
        Initialized `CacheTiles` class.
    """
    write_extension = ".ext"
    write_func_kwargs = {}
    return WriteImage(
        write_func=write_func,
        write_extension=write_extension,
        write_func_kwargs=write_func_kwargs,
    )


def test_cache_tiles_init(write_func, write_image_strategy):
    """
    Test `WriteImage` initializes as expected.
    """
    assert write_image_strategy.write_func is write_func
    assert write_image_strategy.write_extension == ".ext"
    assert write_image_strategy.write_func_kwargs == {}


def test_write_batch(write_image_strategy, ordered_array):

    n_batches = 1

    prediction = ordered_array((n_batches, 1, 8, 8))

    batch = prediction
    batch_indices = np.arange(n_batches)
    batch_idx = 0

    # mock trainer and datasets
    trainer = Mock(spec=Trainer)

    # mock trainer and datasets
    trainer = Mock(spec=Trainer)
    mock_dataset = Mock(spec=IterablePredDataset)
    dataloader_idx = 0
    mock_dl = Mock(spec=DataLoader)
    mock_dl.batch_size = 1
    trainer.predict_dataloaders = [mock_dl]
    trainer.predict_dataloaders[dataloader_idx].dataset = mock_dataset

    # These functions have their own unit tests,
    #   so they do not need to be tested again here.
    # This is a unit test to isolate functionality of `write_batch.`
    with patch.multiple(
        "careamics.lightning.callbacks.prediction_writer_callback.write_strategy",
        get_sample_file_path=DEFAULT,
        create_write_file_path=DEFAULT,
    ) as values:

        # mocked functions
        mock_get_sample_file_path = values["get_sample_file_path"]
        mock_create_write_file_path = values["create_write_file_path"]

        # assign mock functions return value
        in_file_path = Path("in_dir/file_path.ext")
        out_file_path = Path("out_dir/file_path.out_ext")
        mock_get_sample_file_path.return_value = in_file_path
        mock_create_write_file_path.return_value = out_file_path

        # call write batch
        dirpath = "predictions"
        write_image_strategy.write_batch(
            trainer=trainer,
            pl_module=Mock(spec=LightningModule),
            prediction=prediction,
            batch_indices=batch_indices,
            batch=batch,  # contains the last tile
            batch_idx=batch_idx,
            dataloader_idx=dataloader_idx,
            dirpath=dirpath,
        )

        # assert create_write_file_path is called as expected (TODO: necessary ?)
        mock_create_write_file_path.assert_called_with(
            dirpath=dirpath,
            file_path=in_file_path,
            write_extension=write_image_strategy.write_extension,
        )
        # assert write_func is called as expectedÏ
        # cannot use `assert_called_once_with` because of numpy array
        write_image_strategy.write_func.assert_called_once()
        assert (
            write_image_strategy.write_func.call_args.kwargs["file_path"]
            == out_file_path
        )
        assert np.array_equal(
            write_image_strategy.write_func.call_args.kwargs["img"], prediction[0]
        )



from collections.abc import Sequence
from typing import Callable

import numpy as np
import pytest

from careamics.dataset_ng.patching_strategies import (
    FixedRandomPatchingStrategy,
    PatchingStrategy,
    RandomPatchingStrategy,
    SequentialPatchingStrategy,
)


def _create_random_patching_strategy(
    data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
) -> RandomPatchingStrategy:
    seed = 42
    return RandomPatchingStrategy(
        data_shapes=data_shapes, patch_size=patch_size, seed=seed
    )


def _create_fixed_random_patching_strategy(
    data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
) -> FixedRandomPatchingStrategy:
    seed = 42
    return FixedRandomPatchingStrategy(
        data_shapes=data_shapes, patch_size=patch_size, seed=seed
    )


def _create_sequential_patching_strategy(
    data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
) -> SequentialPatchingStrategy:
    overlap = tuple(2 for _ in patch_size)
    return SequentialPatchingStrategy(data_shapes, patch_size, overlap)


PatchingStrategyConstr = Callable[
    [Sequence[Sequence[int]], Sequence[int]], PatchingStrategy
]

# !!! if new strategies are added they should be tested here !!!
PATCHING_STRATEGY_CONSTR: tuple[PatchingStrategyConstr, ...] = (
    _create_random_patching_strategy,
    _create_fixed_random_patching_strategy,
    _create_sequential_patching_strategy,
)


@pytest.mark.parametrize("strategy_constr", PATCHING_STRATEGY_CONSTR)
@pytest.mark.parametrize(
    "data_shapes,patch_size",
    [
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 8)],
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 5)],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 8, 8)],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 5, 7)],
    ],
)
def test_get_all_patch_specs(
    strategy_constr: PatchingStrategyConstr,
    data_shapes: Sequence[Sequence[int]],
    patch_size: Sequence[int],
):
    strategy = strategy_constr(data_shapes, patch_size)
    n_patches = strategy.n_patches
    for i in range(n_patches):
        patch_spec = strategy.get_patch_spec(i)

        # validate patch is within spatial bounds
        coords = np.array(patch_spec["coords"])
        patch_size_ = np.array(patch_spec["patch_size"])
        data_index = patch_spec["data_idx"]
        data_shape = data_shapes[data_index]
        spatial_shape = data_shape[2:]

        assert (0 <= coords).all()
        assert (coords + patch_size_ <= spatial_shape).all()


@pytest.mark.parametrize("strategy_constr", PATCHING_STRATEGY_CONSTR)
@pytest.mark.parametrize(
    "data_shapes,patch_size",
    [
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 8)],
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 5)],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 8, 8)],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 5, 7)],
    ],
)
def test_patches_cover_50percent(
    strategy_constr: PatchingStrategyConstr,
    data_shapes: Sequence[Sequence[int]],
    patch_size: Sequence[int],
):
    # Testing more than 50% because some patching strategies are random (but seeded)
    """Test that more than 50% of the data is covered by the sampled pathches"""
    patching_strategy = strategy_constr(data_shapes, patch_size)

    # track where patches have been sampled from
    tracking_arrays = [np.zeros(data_shape, dtype=bool) for data_shape in data_shapes]

    patch_specs = [
        patching_strategy.get_patch_spec(i) for i in range(patching_strategy.n_patches)
    ]
    for patch_spec in patch_specs:
        tracking_array = tracking_arrays[patch_spec["data_idx"]]
        spatial_slice = tuple(
            slice(c, c + ps)
            for c, ps in zip(patch_spec["coords"], patch_spec["patch_size"])
        )
        # set to true where the patches would be sampled from
        tracking_array[(patch_spec["sample_idx"], slice(None), *spatial_slice)] = True

    total_covered = 0
    total_size = 0
    for tracking_array in tracking_arrays:
        # if the patch specs covered all the image all the values should be true
        total_covered += np.count_nonzero(tracking_array)
        total_size += tracking_array.size
    assert total_covered / total_size > 0.5



"""Type used to represent all transformations users can create."""

from typing import Annotated, Union

from pydantic import Discriminator

from .normalize_model import NormalizeModel
from .xy_flip_model import XYFlipModel
from .xy_random_rotate90_model import XYRandomRotate90Model

NORM_AND_SPATIAL_UNION = Annotated[
    Union[
        NormalizeModel,
        XYFlipModel,
        XYRandomRotate90Model,
    ],
    Discriminator("name"),  # used to tell the different transform models apart
]
"""All transforms including normalization."""


SPATIAL_TRANSFORMS_UNION = Annotated[
    Union[
        XYFlipModel,
        XYRandomRotate90Model,
    ],
    Discriminator("name"),  # used to tell the different transform models apart
]
"""Available spatial transforms in CAREamics."""



import numpy as np
from bioimageio.spec import load_description
from torch import Tensor

from careamics import CAREamist
from careamics.model_io import export_to_bmz, load_pretrained
from careamics.model_io.bmz_io import _export_state_dict, _load_state_dict


def test_state_dict_io(tmp_path, ordered_array, pre_trained):
    """Test exporting and loading a state dict."""
    # training data
    train_array = ordered_array((32, 32))
    path = tmp_path / "model.pth"

    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)

    # predict (no tiling and no tta)
    predicted_output = careamist.predict(train_array, tta_transforms=False)
    predicted = np.concatenate(predicted_output, axis=0)

    # save model
    _export_state_dict(careamist.model, path)
    assert path.exists()

    # load model
    _load_state_dict(careamist.model, path)

    # predict (no tiling and no tta)
    predicted_loaded = careamist.predict(train_array, tta_transforms=False)
    assert (predicted_loaded == predicted).all()


def test_bmz_io(tmp_path, ordered_array, pre_trained):
    """Test exporting and loading to the BMZ."""
    # training data
    train_array = ordered_array((32, 32))

    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)

    # predict (no tiling and no tta)
    predicted_output = careamist.predict(train_array, tta_transforms=False)
    predicted = np.concatenate(predicted_output, axis=0)

    # export to BioImage Model Zoo
    path = tmp_path / "other_folder" / "model.zip"
    export_to_bmz(
        model=careamist.model,
        config=careamist.cfg,
        path_to_archive=path,
        model_name="TopModel",
        general_description="A model that just walked in.",
        data_description="My data.",
        authors=[{"name": "Amod", "affiliation": "El"}],
        input_array=train_array[np.newaxis, np.newaxis, ...],
        output_array=predicted,
        model_version="0.0.15",
    )
    assert path.exists()

    # load description
    description = load_description(path)
    assert str(description.version) == "0.0.15"

    # load model
    model, config = load_pretrained(path)
    assert config == careamist.cfg

    # compare predictions
    torch_array = Tensor(train_array[np.newaxis, np.newaxis, ...])
    predicted = careamist.model.forward(torch_array).detach().numpy().squeeze()
    predicted_loaded = model.forward(torch_array).detach().numpy().squeeze()
    assert (predicted_loaded == predicted).all()



"""Validator utilities."""

__all__ = [
    "check_axes_validity",
    "model_matching_in_out_channels",
    "model_without_final_activation",
    "model_without_n2v2",
    "patch_size_ge_than_8_power_of_2",
]

from .model_validators import (
    model_matching_in_out_channels,
    model_without_final_activation,
    model_without_n2v2,
)
from .validator_utils import check_axes_validity, patch_size_ge_than_8_power_of_2



"""Progressbar callback."""

import sys
from typing import Union

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import TQDMProgressBar
from tqdm.auto import tqdm


class ProgressBarCallback(TQDMProgressBar):
    """Progress bar for training and validation steps."""

    def init_train_tqdm(self) -> tqdm:
        """Override this to customize the tqdm bar for training.

        Returns
        -------
        tqdm
            A tqdm bar.
        """
        bar = tqdm(
            desc="Training",
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=sys.stdout,
            smoothing=0,
        )
        return bar

    def init_validation_tqdm(self) -> tqdm:
        """Override this to customize the tqdm bar for validation.

        Returns
        -------
        tqdm
            A tqdm bar.
        """
        # The main progress bar doesn't exist in `trainer.validate()`
        has_main_bar = self.train_progress_bar is not None
        bar = tqdm(
            desc="Validating",
            position=(2 * self.process_position + has_main_bar),
            disable=self.is_disabled,
            leave=False,
            dynamic_ncols=True,
            file=sys.stdout,
        )
        return bar

    def init_test_tqdm(self) -> tqdm:
        """Override this to customize the tqdm bar for testing.

        Returns
        -------
        tqdm
            A tqdm bar.
        """
        bar = tqdm(
            desc="Testing",
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=False,
            ncols=100,
            file=sys.stdout,
        )
        return bar

    def get_metrics(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> dict[str, Union[int, str, float, dict[str, float]]]:
        """Override this to customize the metrics displayed in the progress bar.

        Parameters
        ----------
        trainer : Trainer
            The trainer object.
        pl_module : LightningModule
            The LightningModule object, unused.

        Returns
        -------
        dict
            A dictionary with the metrics to display in the progress bar.
        """
        pbar_metrics = trainer.progress_bar_metrics
        return {**pbar_metrics}



"""XY flip transform."""

from typing import Optional

import numpy as np
from numpy.typing import NDArray

from careamics.transforms.transform import Transform


class XYFlip(Transform):
    """Flip image along X and Y axis, one at a time.

    This transform randomly flips one of the last two axes.

    This transform expects C(Z)YX dimensions.

    Attributes
    ----------
    axis_indices : List[int]
        Indices of the axes that can be flipped.
    rng : np.random.Generator
        Random number generator.
    p : float
        Probability of applying the transform.
    seed : Optional[int]
        Random seed.

    Parameters
    ----------
    flip_x : bool, optional
        Whether to flip along the X axis, by default True.
    flip_y : bool, optional
        Whether to flip along the Y axis, by default True.
    p : float, optional
        Probability of applying the transform, by default 0.5.
    seed : Optional[int], optional
        Random seed, by default None.
    """

    def __init__(
        self,
        flip_x: bool = True,
        flip_y: bool = True,
        p: float = 0.5,
        seed: Optional[int] = None,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        flip_x : bool, optional
            Whether to flip along the X axis, by default True.
        flip_y : bool, optional
            Whether to flip along the Y axis, by default True.
        p : float
            Probability of applying the transform, by default 0.5.
        seed : Optional[int], optional
            Random seed, by default None.
        """
        if p < 0 or p > 1:
            raise ValueError("Probability must be in [0, 1].")

        if not flip_x and not flip_y:
            raise ValueError("At least one axis must be flippable.")

        # probability to apply the transform
        self.p = p

        # "flippable" axes
        self.axis_indices = []

        if flip_y:
            self.axis_indices.append(-2)
        if flip_x:
            self.axis_indices.append(-1)

        # numpy random generator
        self.rng = np.random.default_rng(seed=seed)

    def __call__(
        self,
        patch: NDArray,
        target: Optional[NDArray] = None,
        **additional_arrays: NDArray,
    ) -> tuple[NDArray, Optional[NDArray], dict[str, NDArray]]:
        """Apply the transform to the source patch and the target (optional).

        Parameters
        ----------
        patch : np.ndarray
            Patch, 2D or 3D, shape C(Z)YX.
        target : Optional[np.ndarray], optional
            Target for the patch, by default None.
        **additional_arrays : NDArray
            Additional arrays that will be transformed identically to `patch` and
            `target`.

        Returns
        -------
        Tuple[np.ndarray, Optional[np.ndarray]]
            Transformed patch and target.
        """
        if self.rng.random() > self.p:
            return patch, target, additional_arrays

        # choose an axis to flip
        axis = self.rng.choice(self.axis_indices)

        patch_transformed = self._apply(patch, axis)
        target_transformed = self._apply(target, axis) if target is not None else None
        additional_transformed = {
            key: self._apply(array, axis) for key, array in additional_arrays.items()
        }

        return patch_transformed, target_transformed, additional_transformed

    def _apply(self, patch: NDArray, axis: int) -> NDArray:
        """Apply the transform to the image.

        Parameters
        ----------
        patch : np.ndarray
            Image patch, 2D or 3D, shape C(Z)YX.
        axis : int
            Axis to flip.

        Returns
        -------
        np.ndarray
            Flipped image patch.
        """
        return np.ascontiguousarray(np.flip(patch, axis=axis))



import pytest

from careamics.config import Configuration


@pytest.mark.parametrize("name", ["Sn4K3", "C4_M e-L"])
def test_valid_names(minimum_supervised_configuration: dict, name: str):
    """Test valid names (letters, numbers, spaces, dashes and underscores)."""
    minimum_supervised_configuration["experiment_name"] = name
    myconf = Configuration(**minimum_supervised_configuration)
    assert myconf.experiment_name == name


@pytest.mark.parametrize("name", ["", "   ", "#", "/", "^", "%", ",", ".", "a=b"])
def test_invalid_names(minimum_supervised_configuration: dict, name: str):
    """Test that invalid names raise an error."""
    minimum_supervised_configuration["experiment_name"] = name
    with pytest.raises(ValueError):
        Configuration(**minimum_supervised_configuration)


def test_3D_algorithm_and_data_compatibility(minimum_supervised_configuration: dict):
    """Test that errors are raised if algorithm `is_3D` and data axes are
    incompatible.
    """
    # 3D but no Z in axes
    minimum_supervised_configuration["algorithm_config"]["model"]["conv_dims"] = 3
    config = Configuration(**minimum_supervised_configuration)
    assert config.algorithm_config.model.conv_dims == 2

    # 2D but Z in axes
    minimum_supervised_configuration["algorithm_config"]["model"]["conv_dims"] = 2
    minimum_supervised_configuration["data_config"]["axes"] = "ZYX"
    minimum_supervised_configuration["data_config"]["patch_size"] = [64, 64, 64]
    config = Configuration(**minimum_supervised_configuration)
    assert config.algorithm_config.model.conv_dims == 3


def test_set_3D(minimum_supervised_configuration: dict):
    """Test the set 3D method."""
    conf = Configuration(**minimum_supervised_configuration)

    # set to 3D
    conf.set_3D(True, "ZYX", [64, 64, 64])
    assert conf.data_config.axes == "ZYX"
    assert conf.data_config.patch_size == [64, 64, 64]
    assert conf.algorithm_config.model.conv_dims == 3

    # set to 2D
    conf.set_3D(False, "SYX", [64, 64])
    assert conf.data_config.axes == "SYX"
    assert conf.data_config.patch_size == [64, 64]
    assert conf.algorithm_config.model.conv_dims == 2



"""Deep-learning model configurations."""

__all__ = ["ArchitectureModel", "LVAEModel", "UNetModel"]

from .architecture_model import ArchitectureModel
from .lvae_model import LVAEModel
from .unet_model import UNetModel



"""Algorithm configurations."""

__all__ = [
    "CAREAlgorithm",
    "N2NAlgorithm",
    "N2VAlgorithm",
    "UNetBasedAlgorithm",
    "VAEBasedAlgorithm",
]

from .care_algorithm_model import CAREAlgorithm
from .n2n_algorithm_model import N2NAlgorithm
from .n2v_algorithm_model import N2VAlgorithm
from .unet_algorithm_model import UNetBasedAlgorithm
from .vae_algorithm_model import VAEBasedAlgorithm



"""Data supported by CAREamics."""

from __future__ import annotations

from typing import Union

from careamics.utils import BaseEnum


class SupportedData(str, BaseEnum):
    """Supported data types.

    Attributes
    ----------
    ARRAY : str
        Array data.
    TIFF : str
        TIFF image data.
    CUSTOM : str
        Custom data.
    """

    ARRAY = "array"
    TIFF = "tiff"
    CUSTOM = "custom"
    # ZARR = "zarr"

    # TODO remove?
    @classmethod
    def _missing_(cls, value: object) -> str:
        """
        Override default behaviour for missing values.

        This method is called when `value` is not found in the enum values. It converts
        `value` to lowercase, removes "." if it is the first character and tries to
        match it with enum values.

        Parameters
        ----------
        value : object
            Value to be matched with enum values.

        Returns
        -------
        str
            Matched enum value.
        """
        if isinstance(value, str):
            lower_value = value.lower()

            if lower_value.startswith("."):
                lower_value = lower_value[1:]

            # attempt to match lowercase value with enum values
            for member in cls:
                if member.value == lower_value:
                    return member

        # still missing
        return super()._missing_(value)

    @classmethod
    def get_extension_pattern(cls, data_type: Union[str, SupportedData]) -> str:
        """
        Get Path.rglob and fnmatch compatible extension.

        Parameters
        ----------
        data_type : SupportedData
            Data type.

        Returns
        -------
        str
            Corresponding extension pattern.
        """
        if data_type == cls.ARRAY:
            raise NotImplementedError(f"Data '{data_type}' is not loaded from a file.")
        elif data_type == cls.TIFF:
            return "*.tif*"
        elif data_type == cls.CUSTOM:
            return "*.*"
        else:
            raise ValueError(f"Data type {data_type} is not supported.")

    @classmethod
    def get_extension(cls, data_type: Union[str, SupportedData]) -> str:
        """
        Get file extension of corresponding data type.

        Parameters
        ----------
        data_type : str or SupportedData
            Data type.

        Returns
        -------
        str
            Corresponding extension.
        """
        if data_type == cls.ARRAY:
            raise NotImplementedError(f"Data '{data_type}' is not loaded from a file.")
        elif data_type == cls.TIFF:
            return ".tiff"
        elif data_type == cls.CUSTOM:
            # TODO: improve this message
            raise NotImplementedError("Custom extensions have to be passed elsewhere.")
        else:
            raise ValueError(f"Data type {data_type} is not supported.")



"""A class chaining transforms together."""

from typing import Optional, Union, cast

from numpy.typing import NDArray

from careamics.config.transformations import NORM_AND_SPATIAL_UNION

from .normalize import Normalize
from .transform import Transform
from .xy_flip import XYFlip
from .xy_random_rotate90 import XYRandomRotate90

ALL_TRANSFORMS = {
    "Normalize": Normalize,
    "XYFlip": XYFlip,
    "XYRandomRotate90": XYRandomRotate90,
}


def get_all_transforms() -> dict[str, type]:
    """Return all the transforms accepted by CAREamics.

    Returns
    -------
    dict
        A dictionary with all the transforms accepted by CAREamics, where the keys are
        the transform names and the values are the transform classes.
    """
    return ALL_TRANSFORMS


class Compose:
    """A class chaining transforms together.

    Parameters
    ----------
    transform_list : list[TransformModel]
        A list of dictionaries where each dictionary contains the name of a
        transform and its parameters.

    Attributes
    ----------
    _callable_transforms : Callable
        A callable that applies the transforms to the input data.
    """

    def __init__(self, transform_list: list[NORM_AND_SPATIAL_UNION]) -> None:
        """Instantiate a Compose object.

        Parameters
        ----------
        transform_list : list[NORM_AND_SPATIAL_UNION]
            A list of dictionaries where each dictionary contains the name of a
            transform and its parameters.
        """
        # retrieve all available transforms
        # TODO: correctly type hint get_all_transforms function output
        all_transforms: dict[str, type[Transform]] = get_all_transforms()

        # instantiate all transforms
        self.transforms: list[Transform] = [
            all_transforms[t.name](**t.model_dump()) for t in transform_list
        ]

    def _chain_transforms(
        self, patch: NDArray, target: Optional[NDArray]
    ) -> tuple[Optional[NDArray], ...]:
        """Chain transforms on the input data.

        Parameters
        ----------
        patch : np.ndarray
            Input data.
        target : Optional[np.ndarray]
            Target data, by default None.

        Returns
        -------
        tuple[np.ndarray, Optional[np.ndarray]]
            The output of the transformations.
        """
        params: Union[tuple[NDArray, Optional[NDArray]],] = (patch, target)

        for t in self.transforms:
            *params, _ = t(*params)  # ignore additional_arrays dict

        # avoid None values that create problems for collating
        return tuple(p for p in params if p is not None)

    def _chain_transforms_additional_arrays(
        self,
        patch: NDArray,
        target: Optional[NDArray],
        **additional_arrays: NDArray,
    ) -> tuple[NDArray, Optional[NDArray], dict[str, NDArray]]:
        """Chain transforms on the input data, with additional arrays.

        Parameters
        ----------
        patch : np.ndarray
            Input data.
        target : Optional[np.ndarray]
            Target data, by default None.
        **additional_arrays : NDArray
            Additional arrays that will be transformed identically to `patch` and
            `target`.

        Returns
        -------
        tuple[np.ndarray, Optional[np.ndarray]]
            The output of the transformations.
        """
        params = {"patch": patch, "target": target, **additional_arrays}

        for t in self.transforms:
            patch, target, additional_arrays = t(**params)
            params = {"patch": patch, "target": target, **additional_arrays}

        return patch, target, additional_arrays

    def __call__(
        self, patch: NDArray, target: Optional[NDArray] = None
    ) -> tuple[NDArray, ...]:
        """Apply the transforms to the input data.

        Parameters
        ----------
        patch : np.ndarray
            The input data.
        target : Optional[np.ndarray], optional
            Target data, by default None.

        Returns
        -------
        tuple[np.ndarray, ...]
            The output of the transformations.
        """
        # TODO: solve casting Compose.__call__ ouput
        return cast(tuple[NDArray, ...], self._chain_transforms(patch, target))

    def transform_with_additional_arrays(
        self,
        patch: NDArray,
        target: Optional[NDArray] = None,
        **additional_arrays: NDArray,
    ) -> tuple[NDArray, Optional[NDArray], dict[str, NDArray]]:
        """Apply the transforms to the input data, including additional arrays.

        Parameters
        ----------
        patch : np.ndarray
            The input data.
        target : Optional[np.ndarray], optional
            Target data, by default None.
        **additional_arrays : NDArray
            Additional arrays that will be transformed identically to `patch` and
            `target`.

        Returns
        -------
        NDArray
            The transformed patch.
        NDArray | None
            The transformed target.
        dict of {str, NDArray}
            Transformed additional arrays. Keys correspond to the keyword argument
            names.
        """
        return self._chain_transforms_additional_arrays(
            patch, target, **additional_arrays
        )



"""Module containing different strategies for writing predictions."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any, Optional, Protocol, Union

import numpy as np
from numpy.typing import NDArray
from pytorch_lightning import LightningModule, Trainer
from torch.utils.data import DataLoader

from careamics.config.tile_information import TileInformation
from careamics.dataset import IterablePredDataset, IterableTiledPredDataset
from careamics.file_io import WriteFunc
from careamics.prediction_utils import stitch_prediction_single

from .file_path_utils import create_write_file_path, get_sample_file_path


class WriteStrategy(Protocol):
    """Protocol for write strategy classes."""

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: Any,  # TODO: change to expected type
        batch_indices: Optional[Sequence[int]],
        batch: Any,  # TODO: change to expected type
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        WriteStrategy subclasses must contain this function to write a batch.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.
        """


class CacheTiles(WriteStrategy):
    """
    A write strategy that will cache tiles.

    Tiles are cached until a whole image is predicted on. Then the stitched
    prediction is saved.

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    tile_cache : list of numpy.ndarray
        Tiles cached for stitching prediction.
    tile_info_cache : list of TileInformation
        Cached tile information for stitching prediction.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A write strategy that will cache tiles.

        Tiles are cached until a whole image is predicted on. Then the stitched
        prediction is saved.

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

        # where tiles will be cached until a whole image has been predicted
        self.tile_cache: list[NDArray] = []
        self.tile_info_cache: list[TileInformation] = []

    @property
    def last_tiles(self) -> list[bool]:
        """
        List of bool to determine whether each tile in the cache is the last tile.

        Returns
        -------
        list of bool
            Whether each tile in the tile cache is the last tile.
        """
        return [tile_info.last_tile for tile_info in self.tile_info_cache]

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: tuple[NDArray, list[TileInformation]],
        batch_indices: Optional[Sequence[int]],
        batch: tuple[NDArray, list[TileInformation]],
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Cache tiles until the last tile is predicted; save the stitched prediction.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.
        """
        dataloaders: Union[DataLoader, list[DataLoader]] = trainer.predict_dataloaders
        dataloader: DataLoader = (
            dataloaders[dataloader_idx]
            if isinstance(dataloaders, list)
            else dataloaders
        )
        dataset: IterableTiledPredDataset = dataloader.dataset
        if not isinstance(dataset, IterableTiledPredDataset):
            raise TypeError("Prediction dataset is not `IterableTiledPredDataset`.")

        # cache tiles (batches are split into single samples)
        self.tile_cache.extend(np.split(prediction[0], prediction[0].shape[0]))
        self.tile_info_cache.extend(prediction[1])

        # save stitched prediction
        if self._has_last_tile():

            # get image tiles and remove them from the cache
            tiles, tile_infos = self._get_image_tiles()
            self._clear_cache()

            # stitch prediction
            prediction_image = stitch_prediction_single(
                tiles=tiles, tile_infos=tile_infos
            )

            # write prediction
            sample_id = tile_infos[0].sample_id  # need this to select correct file name
            input_file_path = get_sample_file_path(dataset=dataset, sample_id=sample_id)
            file_path = create_write_file_path(
                dirpath=dirpath,
                file_path=input_file_path,
                write_extension=self.write_extension,
            )
            self.write_func(
                file_path=file_path, img=prediction_image[0], **self.write_func_kwargs
            )

    def _has_last_tile(self) -> bool:
        """
        Whether a last tile is contained in the cached tiles.

        Returns
        -------
        bool
            Whether a last tile is contained in the cached tiles.
        """
        return any(self.last_tiles)

    def _clear_cache(self) -> None:
        """Remove the tiles in the cache up to the first last tile."""
        index = self._last_tile_index()
        self.tile_cache = self.tile_cache[index + 1 :]
        self.tile_info_cache = self.tile_info_cache[index + 1 :]

    def _last_tile_index(self) -> int:
        """
        Find the index of the last tile in the tile cache.

        Returns
        -------
        int
            Index of last tile.

        Raises
        ------
        ValueError
            If there is no last tile in the tile cache.
        """
        last_tiles = self.last_tiles
        if not any(last_tiles):
            raise ValueError("No last tile in the tile cache.")
        index = np.where(last_tiles)[0][0]
        return index

    def _get_image_tiles(self) -> tuple[list[NDArray], list[TileInformation]]:
        """
        Get the tiles corresponding to a single image.

        Returns
        -------
        tuple of (list of numpy.ndarray, list of TileInformation)
            Tiles and tile information to stitch together a full image.
        """
        index = self._last_tile_index()
        tiles = self.tile_cache[: index + 1]
        tile_infos = self.tile_info_cache[: index + 1]
        return tiles, tile_infos


class WriteTilesZarr(WriteStrategy):
    """Strategy to write tiles to Zarr file."""

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: Any,
        batch_indices: Optional[Sequence[int]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Write tiles to zarr file.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError


class WriteImage(WriteStrategy):
    """
    A strategy for writing image predictions (i.e. un-tiled predictions).

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A strategy for writing image predictions (i.e. un-tiled predictions).

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: NDArray,
        batch_indices: Optional[Sequence[int]],
        batch: NDArray,
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Save full images.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.

        Raises
        ------
        TypeError
            If trainer prediction dataset is not `IterablePredDataset`.
        """
        dls: Union[DataLoader, list[DataLoader]] = trainer.predict_dataloaders
        dl: DataLoader = dls[dataloader_idx] if isinstance(dls, list) else dls
        ds: IterablePredDataset = dl.dataset
        if not isinstance(ds, IterablePredDataset):
            raise TypeError("Prediction dataset is not `IterablePredDataset`.")

        for i in range(prediction.shape[0]):
            prediction_image = prediction[0]
            sample_id = batch_idx * dl.batch_size + i
            input_file_path = get_sample_file_path(dataset=ds, sample_id=sample_id)
            file_path = create_write_file_path(
                dirpath=dirpath,
                file_path=input_file_path,
                write_extension=self.write_extension,
            )
            self.write_func(
                file_path=file_path, img=prediction_image, **self.write_func_kwargs
            )



"""A module to contain type definitions relating to patching strategies."""

from collections.abc import Sequence
from typing import Protocol, TypedDict


class PatchSpecs(TypedDict):
    """A dictionary that specifies a single patch in a series of `ImageStacks`.

    Attributes
    ----------
    data_idx: int
        Determines which `ImageStack` a patch belongs to, within a series of
        `ImageStack`s.
    sample_idx: int
        Determines which sample a patch belongs to, within an `ImageStack`.
    coords: sequence of int
        The top-left (and first z-slice for 3D data) of a patch. The sequence will have
        length 2 or 3, for 2D and 3D data respectively.
    patch_size: sequence of int
        The size of the patch. The sequence will have length 2 or 3, for 2D and 3D data
        respectively.
    """

    data_idx: int
    sample_idx: int
    coords: Sequence[int]
    patch_size: Sequence[int]


class PatchingStrategy(Protocol):
    """
    An interface for patching strategies.

    Patching strategies are a component of the `CAREamicsDataset`; they determine
    how patches are extracted from the underlying data.

    Attributes
    ----------
    n_patches: int
        The number of patches that the patching strategy will return.

    Methods
    -------
    get_patch_spec(index: int) -> PatchSpecs
        Get a patch specification for a given patch index.
    """

    @property
    def n_patches(self) -> int:
        """
        The number of patches that the patching strategy will return.

        It also determines the maximum index that can be given to `get_patch_spec`,
        and the length of the `CAREamicsDataset`.

        Returns
        -------
        int
            Number of patches.
        """
        ...

    def get_patch_spec(self, index: int) -> PatchSpecs:
        """
        Get a patch specification for a given patch index.

        This method is intended to be called from within the
        `CAREamicsDataset.__getitem__`. The index will be passed through from this
        method.

        Parameters
        ----------
        index : int
            A patch index.

        Returns
        -------
        PatchSpecs
            A dictionary that specifies a single patch in a series of `ImageStacks`.
        """
        ...



from collections.abc import Sequence
from pathlib import Path
from typing import Literal, Protocol, Union

from numpy.typing import DTypeLike, NDArray


class ImageStack(Protocol):
    """
    An interface for extracting patches from an image stack.

    Attributes
    ----------
    source: Path or "array"
        Origin of the image data.
    data_shape: Sequence[int]
        The shape of the data, it is expected to be in the order (SC(Z)YX).

    """

    # TODO: not sure how compatible using Path will be for a zarr array
    #   (for a zarr array need to specify file path and internal zarr path)
    @property
    def source(self) -> Union[Path, Literal["array"]]: ...

    @property
    def data_shape(self) -> Sequence[int]: ...

    @property
    def data_dtype(self) -> DTypeLike: ...

    def extract_patch(
        self, sample_idx: int, coords: Sequence[int], patch_size: Sequence[int]
    ) -> NDArray:
        """
        Extracts a patch for a given sample within the image stack.

        Parameters
        ----------
        sample_idx: int
            Sample index. The first dimension of the image data will be indexed at this
            value.
        coords: Sequence of int
            The coordinates that define the start of a patch.
        patch_size: Sequence of int
            The size of the patch in each spatial dimension.

        Returns
        -------
        numpy.ndarray
            A patch of the image data from a particlular sample. It will have the
            dimensions C(Z)YX.
        """



"""UNet Pydantic model."""

from __future__ import annotations

from typing import Literal

from pydantic import ConfigDict, Field, field_validator

from .architecture_model import ArchitectureModel


# TODO tests activation <-> pydantic model, test the literals!
# TODO annotations for the json schema?
class UNetModel(ArchitectureModel):
    """
    Pydantic model for a N2V(2)-compatible UNet.

    Attributes
    ----------
    depth : int
        Depth of the model, between 1 and 10 (default 2).
    num_channels_init : int
        Number of filters of the first level of the network, should be even
        and minimum 8 (default 96).
    """

    # pydantic model config
    model_config = ConfigDict(validate_assignment=True)

    # discriminator used for choosing the pydantic model in Model
    architecture: Literal["UNet"]
    """Name of the architecture."""

    # parameters
    # validate_defaults allow ignoring default values in the dump if they were not set
    conv_dims: Literal[2, 3] = Field(default=2, validate_default=True)
    """Dimensions (2D or 3D) of the convolutional layers."""

    num_classes: int = Field(default=1, ge=1, validate_default=True)
    """Number of classes or channels in the model output."""

    in_channels: int = Field(default=1, ge=1, validate_default=True)
    """Number of channels in the input to the model."""

    depth: int = Field(default=2, ge=1, le=10, validate_default=True)
    """Number of levels in the UNet."""

    num_channels_init: int = Field(default=32, ge=8, le=1024, validate_default=True)
    """Number of convolutional filters in the first layer of the UNet."""

    # TODO we are not using this, so why make it a choice?
    final_activation: Literal[
        "None", "Sigmoid", "Softmax", "Tanh", "ReLU", "LeakyReLU"
    ] = Field(default="None", validate_default=True)
    """Final activation function."""

    n2v2: bool = Field(default=False, validate_default=True)
    """Whether to use N2V2 architecture modifications, with blur pool layers and fewer
    skip connections.
    """

    independent_channels: bool = Field(default=True, validate_default=True)
    """Whether information is processed independently in each channel, used to train
    channels independently."""

    @field_validator("num_channels_init")
    @classmethod
    def validate_num_channels_init(cls, num_channels_init: int) -> int:
        """
        Validate that num_channels_init is even.

        Parameters
        ----------
        num_channels_init : int
            Number of channels.

        Returns
        -------
        int
            Validated number of channels.

        Raises
        ------
        ValueError
            If the number of channels is odd.
        """
        # if odd
        if num_channels_init % 2 != 0:
            raise ValueError(
                f"Number of channels for the bottom layer must be even"
                f" (got {num_channels_init})."
            )

        return num_channels_init

    def set_3D(self, is_3D: bool) -> None:
        """
        Set 3D model by setting the `conv_dims` parameters.

        Parameters
        ----------
        is_3D : bool
            Whether the algorithm is 3D or not.
        """
        if is_3D:
            self.conv_dims = 3
        else:
            self.conv_dims = 2

    def is_3D(self) -> bool:
        """
        Return whether the model is 3D or not.

        Returns
        -------
        bool
            Whether the model is 3D or not.
        """
        return self.conv_dims == 3



import numpy as np
import pytest
import torch
from torch.utils.data import DataLoader

from careamics.config.inference_model import InferenceConfig
from careamics.config.tile_information import TileInformation
from careamics.dataset import InMemoryTiledPredDataset
from careamics.dataset.tiling.collate_tiles import collate_tiles
from careamics.models.lvae.likelihoods import GaussianLikelihood
from careamics.models.lvae.lvae import LadderVAE
from careamics.prediction_utils import convert_outputs
from careamics.prediction_utils.lvae_prediction import (
    lvae_predict_mmse_tiled_batch,
    lvae_predict_tiled_batch,
)

# TODO: Test with mock LCMultiChDloader


@pytest.mark.skip(
    reason=(
        "Doesn't make sense to use `InMemoryTiledPredDataset` dataset because it does "
        "not handle lateral context inputs."
    )
)
@pytest.mark.parametrize("predict_logvar", ["pixelwise", None])
@pytest.mark.parametrize("output_channels", [2, 3])
def test_smoke_careamics_dset_lvae_prediction(
    minimum_lvae_params, gaussian_likelihood_params, predict_logvar, output_channels
):
    minimum_lvae_params["predict_logvar"] = predict_logvar
    minimum_lvae_params["output_channels"] = output_channels
    gaussian_likelihood_params["predict_logvar"] = predict_logvar
    input_shape = minimum_lvae_params["input_shape"]

    # initialize model
    model = LadderVAE(**minimum_lvae_params)
    # initialize likelihood
    likelihood_obj = GaussianLikelihood(**gaussian_likelihood_params)

    # create predict dataset
    inference_dict = {
        "data_type": "array",
        "axes": "SYX",
        "tile_size": [input_shape, input_shape],
        "tile_overlap": [(input_shape // 4) * 2, (input_shape // 4) * 2],  # ensure even
        "image_means": [2.0],
        "image_stds": [1.0],
        "tta_transforms": False,
    }
    inference_config = InferenceConfig(**inference_dict)
    N_samples = 3
    data_shape = (N_samples, input_shape * 4 + 23, input_shape * 4 + 23)
    data = np.random.random_sample(size=data_shape)
    dataset = InMemoryTiledPredDataset(inference_config, data)
    dataloader = DataLoader(dataset, collate_fn=collate_tiles)

    tiled_predictions = []
    log_vars = []
    for batch in dataloader:
        y, log_var = lvae_predict_tiled_batch(model, likelihood_obj, batch)
        tiled_predictions.append(y)
        log_vars.append(log_var)

        # y is a 2-tuple, second element is tile info, similar for logvar
        assert y[0].shape == (1, output_channels, input_shape, input_shape)
        if predict_logvar == "pixelwise":
            assert log_var[0].shape == (1, output_channels, input_shape, input_shape)
        elif predict_logvar is None:
            assert log_var is None

    prediction_shape = (1, output_channels, *data_shape[-2:])
    predictions = convert_outputs(tiled_predictions, tiled=True)
    for prediction in predictions:
        assert prediction.shape == prediction_shape


@pytest.mark.parametrize("predict_logvar", ["pixelwise", None])
@pytest.mark.parametrize("output_channels", [2, 3])
def test_lvae_predict_single_sample(
    minimum_lvae_params, gaussian_likelihood_params, predict_logvar, output_channels
):
    """Test predictions of a single sample."""
    minimum_lvae_params["predict_logvar"] = predict_logvar
    minimum_lvae_params["output_channels"] = output_channels
    gaussian_likelihood_params["predict_logvar"] = predict_logvar

    input_shape = minimum_lvae_params["input_shape"]

    # initialize model
    model = LadderVAE(**minimum_lvae_params)
    # initialize likelihood
    likelihood_obj = GaussianLikelihood(**gaussian_likelihood_params)

    # dummy input
    x = torch.rand(size=(1, 1, *input_shape))
    tile_info = TileInformation(
        array_shape=(1, input_shape[0] * 4, input_shape[1] * 4),
        last_tile=False,
        overlap_crop_coords=((8, 8 + input_shape[0]), (8, 8 + input_shape[1])),
        stitch_coords=((0, input_shape[0]), (0, input_shape[1])),
        sample_id=0,
    )
    input_ = (x, [tile_info])  # simulate output of datasets
    # prediction
    y_tiled, log_var_tiled = lvae_predict_tiled_batch(model, likelihood_obj, input_)
    y = y_tiled[0]

    assert y.shape == (1, output_channels, *input_shape)
    if predict_logvar == "pixelwise":
        log_var = log_var_tiled[0]
        assert log_var.shape == (1, output_channels, *input_shape)
    elif predict_logvar is None:
        assert log_var_tiled is None


@pytest.mark.parametrize("predict_logvar", ["pixelwise", None])
@pytest.mark.parametrize("output_channels", [2, 3])
def test_lvae_predict_mmse_tiled_batch(
    minimum_lvae_params, gaussian_likelihood_params, predict_logvar, output_channels
):
    """Test MMSE prediction."""
    minimum_lvae_params["predict_logvar"] = predict_logvar
    minimum_lvae_params["output_channels"] = output_channels
    gaussian_likelihood_params["predict_logvar"] = predict_logvar

    input_shape = minimum_lvae_params["input_shape"]

    # initialize model
    model = LadderVAE(**minimum_lvae_params)
    # initialize likelihood
    likelihood_obj = GaussianLikelihood(**gaussian_likelihood_params)

    # dummy input
    x = torch.rand(size=(1, 1, *input_shape))
    tile_info = TileInformation(
        array_shape=(1, input_shape[0] * 4, input_shape[1] * 4),
        last_tile=False,
        overlap_crop_coords=((8, 8 + input_shape[0]), (8, 8 + input_shape[1])),
        stitch_coords=((0, input_shape[0]), (0, input_shape[1])),
        sample_id=0,
    )  # TODO add 3D case
    input_ = (x, [tile_info])  # simulate output of datasets
    # prediction
    y_tiled, y_std_tiled, log_var_tiled = lvae_predict_mmse_tiled_batch(
        model, likelihood_obj, input_, mmse_count=5
    )
    y = y_tiled[0]
    y_std = y_std_tiled[0]

    assert y.shape == (1, output_channels, *input_shape)
    assert y_std.shape == (1, output_channels, *input_shape)
    if predict_logvar == "pixelwise":
        log_var = log_var_tiled[0]
        assert log_var.shape == (1, output_channels, *input_shape)
    elif predict_logvar is None:
        assert log_var_tiled is None



import numpy as np
import pytest
from tifffile import imwrite

from careamics.config import DataConfig
from careamics.config.support import (
    SupportedData,
    SupportedTransform,
)
from careamics.config.transformations import N2VManipulateModel, XYFlipModel
from careamics.dataset import InMemoryDataset, PathIterableDataset
from careamics.lightning import TrainDataModule, create_train_datamodule


@pytest.fixture
def simple_array(ordered_array):
    return ordered_array((10, 10))


def test_mismatching_types_array(simple_array, minimum_algorithm_n2v):
    """Test that an error is raised if the data type does not match the passed data."""
    minimum_algorithm_n2v["data_type"] = SupportedData.TIFF.value
    with pytest.raises(ValueError):
        TrainDataModule(
            data_config=DataConfig(**minimum_algorithm_n2v), train_data=simple_array
        )

    minimum_algorithm_n2v["data_type"] = SupportedData.CUSTOM.value
    with pytest.raises(ValueError):
        TrainDataModule(
            data_config=DataConfig(**minimum_algorithm_n2v), train_data=simple_array
        )

    minimum_algorithm_n2v["data_type"] = SupportedData.ARRAY.value
    with pytest.raises(ValueError):
        TrainDataModule(
            data_config=DataConfig(**minimum_algorithm_n2v), train_data="path/to/data"
        )


def test_wrapper_unknown_type(simple_array):
    """Test that an error is raised if the data type is not supported."""
    with pytest.raises(ValueError):
        create_train_datamodule(
            train_data=simple_array,
            data_type="wrong_type",
            patch_size=(8, 8),
            axes="YX",
            batch_size=2,
        )


def test_wrapper_train_array(simple_array):
    """Test that the data module is created correctly with an array."""
    # create data module
    data_module = create_train_datamodule(
        train_data=simple_array,
        data_type="array",
        patch_size=(8, 8),
        axes="YX",
        batch_size=2,
        val_minimum_patches=2,
    )
    data_module.prepare_data()
    data_module.setup()

    assert len(list(data_module.train_dataloader())) > 0


def test_wrapper_supervised(simple_array):
    """Test that a supervised data config is created."""
    data_module = create_train_datamodule(
        train_data=simple_array,
        data_type="array",
        patch_size=(8, 8),
        axes="YX",
        batch_size=2,
        train_target_data=simple_array,
        val_minimum_patches=2,
    )
    for transform in data_module.data_config.transforms:
        assert transform.name != SupportedTransform.N2V_MANIPULATE.value


def test_wrapper_supervised_n2v_throws_error(simple_array):
    """Test that an error is raised if target data is passed but the transformations
    (default ones) contain N2V manipulate."""
    with pytest.raises(ValueError):
        create_train_datamodule(
            train_data=simple_array,
            data_type="array",
            patch_size=(8, 8),
            axes="YX",
            batch_size=2,
            train_target_data=simple_array,
            val_minimum_patches=2,
            transforms=[XYFlipModel(), N2VManipulateModel()],
        )


def test_get_data_statistics(tmp_path):
    """Test that get_data_statistics works for every type of training dataset."""
    rng = np.random.default_rng(42)
    data = rng.integers(0, 255, (5, 3, 32, 32))
    data_mean = data.mean(axis=(0, 2, 3))
    data_std = data.std(axis=(0, 2, 3))

    data_val = rng.integers(0, 10, (2, 3, 32, 32))

    # create data module with in memory
    data_config = DataConfig(
        data_type=SupportedData.ARRAY.value,
        patch_size=(16, 16),
        axes="SCYX",
        batch_size=1,
    )
    data_module = TrainDataModule(
        data_config=data_config,
        train_data=data,
        val_data=data_val,
    )
    data_module.prepare_data()
    data_module.setup()
    assert isinstance(data_module.train_dataset, InMemoryDataset)

    means, stds = data_module.get_data_statistics()
    assert np.allclose(means, data_mean)
    assert np.allclose(stds, data_std)

    # same for path iterable
    train_path = tmp_path / "train"
    train_path.mkdir()
    val_path = tmp_path / "val"
    val_path.mkdir()

    for i in range(data.shape[0]):
        data_path = train_path / f"data_{i}.tif"
        imwrite(data_path, data[i])
    for i in range(data_val.shape[0]):
        data_path = val_path / f"data_{i}.tif"
        imwrite(data_path, data_val[i])

    data_config = DataConfig(
        data_type=SupportedData.TIFF.value,
        patch_size=(16, 16),
        axes="CYX",
        batch_size=1,
    )
    data_module = TrainDataModule(
        data_config=data_config,
        train_data=train_path,
        val_data=val_path,
        use_in_memory=False,
    )
    data_module.prepare_data()
    data_module.setup()
    assert isinstance(data_module.train_dataset, PathIterableDataset)

    means, stds = data_module.get_data_statistics()
    assert np.allclose(means, data_mean)
    assert np.allclose(stds, data_std)



# %%
import numpy as np

from careamics.config import create_n2n_configuration
from careamics.config.support import SupportedData
from careamics.dataset_ng.patch_extractor.patch_extractor_factory import (
    create_patch_extractors,
)

rng = np.random.default_rng()

# %%
# define example data
array1 = np.arange(36).reshape(1, 6, 6)
array2 = np.arange(50).reshape(2, 5, 5)
target1 = rng.integers(0, 1, size=array1.shape, endpoint=True)
target2 = rng.integers(0, 1, size=array2.shape, endpoint=True)

# %%
config = create_n2n_configuration(
    "test_exp",
    data_type="array",
    axes="SYX",
    patch_size=[8, 8],
    batch_size=1,
    num_epochs=1,
)
data_config = config.data_config

# %%
data_type = SupportedData(data_config.data_type)
train_inputs, train_targets = create_patch_extractors(
    [array1, array2], [target1, target2], axes=data_config.axes, data_type=data_type
)

# %%
train_inputs.extract_patch(data_idx=0, sample_idx=0, coords=(2, 2), patch_size=(3, 3))



"""Callback saving CAREamics configuration as hyperparameters in the model."""

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import Callback

from careamics.config import Configuration


class HyperParametersCallback(Callback):
    """
    Callback allowing saving CAREamics configuration as hyperparameters in the model.

    This allows saving the configuration as dictionary in the checkpoints, and
    loading it subsequently in a CAREamist instance.

    Parameters
    ----------
    config : Configuration
        CAREamics configuration to be saved as hyperparameter in the model.

    Attributes
    ----------
    config : Configuration
        CAREamics configuration to be saved as hyperparameter in the model.
    """

    def __init__(self, config: Configuration) -> None:
        """
        Constructor.

        Parameters
        ----------
        config : Configuration
            CAREamics configuration to be saved as hyperparameter in the model.
        """
        self.config = config

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """
        Update the hyperparameters of the model with the configuration on train start.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning trainer, unused.
        pl_module : LightningModule
            PyTorch Lightning module.
        """
        pl_module.hparams.update(self.config.model_dump())



"""
Script containing modules for defining different likelihood functions (as nn.Module).
"""

from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import numpy as np
import torch
from torch import nn

from careamics.config.likelihood_model import (
    GaussianLikelihoodConfig,
    NMLikelihoodConfig,
)

if TYPE_CHECKING:
    from careamics.models.lvae.noise_models import (
        GaussianMixtureNoiseModel,
        MultiChannelNoiseModel,
    )

    NoiseModel = Union[GaussianMixtureNoiseModel, MultiChannelNoiseModel]


def likelihood_factory(
    config: Optional[Union[GaussianLikelihoodConfig, NMLikelihoodConfig]],
    noise_model: Optional[NoiseModel] = None,
):
    """
    Factory function for creating likelihood modules.

    Parameters
    ----------
    config: Union[GaussianLikelihoodConfig, NMLikelihoodConfig]
        The configuration object for the likelihood module.
    noise_model: Optional[NoiseModel]
        The noise model instance used to define the `NoiseModelLikelihood`.

    Returns
    -------
    nn.Module
        The likelihood module.
    """
    if config is None:
        return None

    if isinstance(config, GaussianLikelihoodConfig):
        return GaussianLikelihood(
            predict_logvar=config.predict_logvar,
            logvar_lowerbound=config.logvar_lowerbound,
        )
    elif isinstance(config, NMLikelihoodConfig):
        return NoiseModelLikelihood(
            data_mean=config.data_mean,
            data_std=config.data_std,
            noise_model=noise_model,
        )
    else:
        raise ValueError(f"Invalid likelihood model type: {config.model_type}")


# TODO: is it really worth to have this class? Or it just adds complexity? --> REFACTOR
class LikelihoodModule(nn.Module):
    """
    The base class for all likelihood modules.
    It defines the fundamental structure and methods for specialized likelihood models.
    """

    def distr_params(self, x: Any) -> None:
        return None

    def set_params_to_same_device_as(self, correct_device_tensor: Any) -> None:
        pass

    @staticmethod
    def logvar(params: Any) -> None:
        return None

    @staticmethod
    def mean(params: Any) -> None:
        return None

    @staticmethod
    def mode(params: Any) -> None:
        return None

    @staticmethod
    def sample(params: Any) -> None:
        return None

    def log_likelihood(self, x: Any, params: Any) -> None:
        return None

    def get_mean_lv(
        self, x: torch.Tensor
    ) -> tuple[torch.Tensor, Optional[torch.Tensor]]: ...

    def forward(
        self, input_: torch.Tensor, x: Union[torch.Tensor, None]
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """
        Parameters
        ----------
        input_: torch.Tensor
            The output of the top-down pass (e.g., reconstructed image in HDN,
            or the unmixed images in 'Split' models).
        x: Union[torch.Tensor, None]
            The target tensor. If None, the log-likelihood is not computed.
        """
        distr_params = self.distr_params(input_)
        mean = self.mean(distr_params)
        mode = self.mode(distr_params)
        sample = self.sample(distr_params)
        logvar = self.logvar(distr_params)

        if x is None:
            ll = None
        else:
            ll = self.log_likelihood(x, distr_params)

        dct = {
            "mean": mean,
            "mode": mode,
            "sample": sample,
            "params": distr_params,
            "logvar": logvar,
        }

        return ll, dct


class GaussianLikelihood(LikelihoodModule):
    r"""A specialized `LikelihoodModule` for Gaussian likelihood.

    Specifically, in the LVAE model, the likelihood is defined as:
        p(x|z_1) = N(x|\mu_{p,1}, \sigma_{p,1}^2)
    """

    def __init__(
        self,
        predict_logvar: Union[Literal["pixelwise"], None] = None,
        logvar_lowerbound: Union[float, None] = None,
    ):
        """Constructor.

        Parameters
        ----------
        predict_logvar: Union[Literal["pixelwise"], None], optional
            If `pixelwise`, log-variance is computed for each pixel, else log-variance
            is not computed. Default is `None`.
        logvar_lowerbound: float, optional
            The lowerbound value for log-variance. Default is `None`.
        """
        super().__init__()

        self.predict_logvar = predict_logvar
        self.logvar_lowerbound = logvar_lowerbound
        assert self.predict_logvar in [None, "pixelwise"]

        print(
            f"[{self.__class__.__name__}] PredLVar:{self.predict_logvar} LowBLVar:{self.logvar_lowerbound}"
        )

    def get_mean_lv(
        self, x: torch.Tensor
    ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Given the output of the top-down pass, compute the mean and log-variance of the
        Gaussian distribution defining the likelihood.

        Parameters
        ----------
        x: torch.Tensor
            The input tensor to the likelihood module, i.e., the output of the top-down
            pass.

        Returns
        -------
        tuple of (torch.tensor, optional torch.tensor)
            The first element of the tuple is the mean, the second element is the
            log-variance. If the attribute `predict_logvar` is `None` then the second
            element will be `None`.
        """
        # if LadderVAE.predict_logvar is None, dim 1 of `x`` has no. of target channels
        if self.predict_logvar is None:
            return x, None

        # Get pixel-wise mean and logvar
        # if LadderVAE.predict_logvar is not None,
        #   dim 1 has double no. of target channels
        mean, lv = x.chunk(2, dim=1)

        # Optionally, clip log-var to a lower bound
        if self.logvar_lowerbound is not None:
            lv = torch.clip(lv, min=self.logvar_lowerbound)

        return mean, lv

    def distr_params(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        """
        Get parameters (mean, log-var) of the Gaussian distribution defined by the likelihood.

        Parameters
        ----------
        x: torch.Tensor
            The input tensor to the likelihood module, i.e., the output
            the LVAE 'output_layer'. Shape is: (B, 2 * C, [Z], Y, X) in case
            `predict_logvar` is not None, or (B, C, [Z], Y, X) otherwise.
        """
        mean, lv = self.get_mean_lv(x)
        params = {
            "mean": mean,
            "logvar": lv,
        }
        return params

    @staticmethod
    def mean(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["mean"]

    @staticmethod
    def mode(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["mean"]

    @staticmethod
    def sample(params: dict[str, torch.Tensor]) -> torch.Tensor:
        # p = Normal(params['mean'], (params['logvar'] / 2).exp())
        # return p.rsample()
        return params["mean"]

    @staticmethod
    def logvar(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["logvar"]

    def log_likelihood(
        self, x: torch.Tensor, params: dict[str, Union[torch.Tensor, None]]
    ):
        """Compute Gaussian log-likelihood

        Parameters
        ----------
        x: torch.Tensor
            The target tensor. Shape is (B, C, [Z], Y, X).
        params: dict[str, Union[torch.Tensor, None]]
            The tensors obtained by chunking the output of the top-down pass,
            here used as parameters of the Gaussian distribution.

        Returns
        -------
        torch.Tensor
            The log-likelihood tensor. Shape is (B, C, [Z], Y, X).
        """
        if self.predict_logvar is not None:
            logprob = log_normal(x, params["mean"], params["logvar"])
        else:
            logprob = -0.5 * (params["mean"] - x) ** 2
        return logprob


def log_normal(
    x: torch.Tensor, mean: torch.Tensor, logvar: torch.Tensor
) -> torch.Tensor:
    """
    Compute the log-probability at `x` of a Gaussian distribution
    with parameters `(mean, exp(logvar))`.

    NOTE: In the case of LVAE, the log-likeihood formula becomes:
        \\mathbb{E}_{z_1\\sim{q_\\phi}}[\\log{p_\theta(x|z_1)}]=-\frac{1}{2}(\\mathbb{E}_{z_1\\sim{q_\\phi}}[\\log{2\\pi\\sigma_{p,0}^2(z_1)}] +\\mathbb{E}_{z_1\\sim{q_\\phi}}[\frac{(x-\\mu_{p,0}(z_1))^2}{\\sigma_{p,0}^2(z_1)}])

    Parameters
    ----------
    x: torch.Tensor
        The ground-truth tensor. Shape is (batch, channels, dim1, dim2).
    mean: torch.Tensor
        The inferred mean of distribution. Shape is (batch, channels, dim1, dim2).
    logvar: torch.Tensor
        The inferred log-variance of distribution. Shape has to be either scalar or broadcastable.
    """
    var = torch.exp(logvar)
    log_prob = -0.5 * (
        ((x - mean) ** 2) / var + logvar + torch.tensor(2 * math.pi).log()
    )
    return log_prob


class NoiseModelLikelihood(LikelihoodModule):

    def __init__(
        self,
        data_mean: Union[np.ndarray, torch.Tensor],
        data_std: Union[np.ndarray, torch.Tensor],
        noise_model: NoiseModel,
    ):
        """Constructor.

        Parameters
        ----------
        data_mean: Union[np.ndarray, torch.Tensor]
            The mean of the data, used to unnormalize data for noise model evaluation.
        data_std: Union[np.ndarray, torch.Tensor]
            The standard deviation of the data, used to unnormalize data for noise
            model evaluation.
        noiseModel: NoiseModel
            The noise model instance used to compute the likelihood.
        """
        super().__init__()
        self.data_mean = torch.Tensor(data_mean)
        self.data_std = torch.Tensor(data_std)
        self.noiseModel = noise_model

    def _set_params_to_same_device_as(
        self, correct_device_tensor: torch.Tensor
    ) -> None:
        """Set the parameters to the same device as the input tensor.

        Parameters
        ----------
        correct_device_tensor: torch.Tensor
            The tensor whose device is used to set the parameters.
        """
        if self.data_mean.device != correct_device_tensor.device:
            self.data_mean = self.data_mean.to(correct_device_tensor.device)
            self.data_std = self.data_std.to(correct_device_tensor.device)
        if correct_device_tensor.device != self.noiseModel.device:
            self.noiseModel.to_device(correct_device_tensor.device)

    def get_mean_lv(self, x: torch.Tensor) -> tuple[torch.Tensor, None]:
        return x, None

    def distr_params(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        mean, lv = self.get_mean_lv(x)
        params = {
            "mean": mean,
            "logvar": lv,
        }
        return params

    @staticmethod
    def mean(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["mean"]

    @staticmethod
    def mode(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["mean"]

    @staticmethod
    def sample(params: dict[str, torch.Tensor]) -> torch.Tensor:
        return params["mean"]

    def log_likelihood(self, x: torch.Tensor, params: dict[str, torch.Tensor]):
        """Compute the log-likelihood given the parameters `params` obtained
        from the reconstruction tensor and the target tensor `x`.

        Parameters
        ----------
        x: torch.Tensor
            The target tensor. Shape is (B, C, [Z], Y, X).
        params: dict[str, Union[torch.Tensor, None]]
            The tensors obtained from output of the top-down pass.
            Here, "mean" correspond to the whole output, while logvar is `None`.

        Returns
        -------
        torch.Tensor
            The log-likelihood tensor. Shape is (B, C, [Z], Y, X).
        """
        self._set_params_to_same_device_as(x)
        predicted_s_denormalized = params["mean"] * self.data_std + self.data_mean
        x_denormalized = x * self.data_std + self.data_mean
        likelihoods = self.noiseModel.likelihood(
            x_denormalized, predicted_s_denormalized
        )
        logprob = torch.log(likelihoods)
        return logprob



"""Function to read zarr images."""

from typing import Union

from zarr import Group, core, hierarchy, storage


def read_zarr(
    zarr_source: Group, axes: str
) -> Union[core.Array, storage.DirectoryStore, hierarchy.Group]:
    """Read a file and returns a pointer.

    Parameters
    ----------
    zarr_source : Group
        Zarr storage.
    axes : str
        Axes of the data.

    Returns
    -------
    np.ndarray
        Pointer to zarr storage.

    Raises
    ------
    ValueError, OSError
        if a file is not a valid tiff or damaged.
    ValueError
        if data dimensions are not 2, 3 or 4.
    ValueError
        if axes parameter from config is not consistent with data dimensions.
    """
    if isinstance(zarr_source, hierarchy.Group):
        array = zarr_source[0]

    elif isinstance(zarr_source, storage.DirectoryStore):
        raise NotImplementedError("DirectoryStore not supported yet")

    elif isinstance(zarr_source, core.Array):
        # array should be of shape (S, (C), (Z), Y, X), iterating over S ?
        if zarr_source.dtype == "O":
            raise NotImplementedError("Object type not supported yet")
        else:
            array = zarr_source
    else:
        raise ValueError(f"Unsupported zarr object type {type(zarr_source)}")

    # sanity check on dimensions
    if len(array.shape) < 2 or len(array.shape) > 4:
        raise ValueError(
            f"Incorrect data dimensions. Must be 2, 3 or 4 (got {array.shape})."
        )

    # sanity check on axes length
    if len(axes) != len(array.shape):
        raise ValueError(f"Incorrect axes length (got {axes}).")

    # arr = fix_axes(arr, axes)
    return array



"""Pydantic model for the N2VManipulate transform."""

from typing import Literal

from pydantic import ConfigDict, Field, field_validator

from .transform_model import TransformModel


# TODO should probably not be a TransformModel anymore, no reason for it
# `name` is used as a discriminator field in the transforms
class N2VManipulateModel(TransformModel):
    """
    Pydantic model used to represent N2V manipulation.

    Attributes
    ----------
    name : Literal["N2VManipulate"]
        Name of the transformation.
    roi_size : int
        Size of the masking region, by default 11.
    masked_pixel_percentage : float
        Percentage of masked pixels, by default 0.2.
    strategy : Literal["uniform", "median"]
        Strategy pixel value replacement, by default "uniform".
    struct_mask_axis : Literal["horizontal", "vertical", "none"]
        Axis of the structN2V mask, by default "none".
    struct_mask_span : int
        Span of the structN2V mask, by default 5.
    """

    model_config = ConfigDict(
        validate_assignment=True,
    )

    name: Literal["N2VManipulate"] = "N2VManipulate"

    roi_size: int = Field(default=11, ge=3, le=21)
    """Size of the region where the pixel manipulation is applied."""

    masked_pixel_percentage: float = Field(default=0.2, ge=0.05, le=10.0)
    """Percentage of masked pixels per image."""

    remove_center: bool = Field(default=True)  # TODO remove it
    """Exclude center pixel from average calculation."""  # TODO rephrase this

    strategy: Literal["uniform", "median"] = Field(default="uniform")
    """Strategy for pixel value replacement."""

    struct_mask_axis: Literal["horizontal", "vertical", "none"] = Field(default="none")
    """Orientation of the structN2V mask. Set to `\"non\"` to not apply StructN2V."""

    struct_mask_span: int = Field(default=5, ge=3, le=15)
    """Size of the structN2V mask."""

    @field_validator("roi_size", "struct_mask_span")
    @classmethod
    def odd_value(cls, v: int) -> int:
        """
        Validate that the value is odd.

        Parameters
        ----------
        v : int
            Value to validate.

        Returns
        -------
        int
            The validated value.

        Raises
        ------
        ValueError
            If the value is even.
        """
        if v % 2 == 0:
            raise ValueError("Size must be an odd number.")
        return v



import pytest

from careamics.config import InferenceConfig
from careamics.config.support import SupportedData
from careamics.lightning import PredictDataModule, create_predict_datamodule


@pytest.fixture
def simple_array(ordered_array):
    return ordered_array((10, 10))


def test_mismatching_types_array(simple_array, minimum_algorithm_n2v):
    """Test that an error is raised if the data type does not match the passed data."""
    minimum_algorithm_n2v["data_type"] = SupportedData.TIFF.value
    with pytest.raises(ValueError):
        PredictDataModule(
            data_config=InferenceConfig(**minimum_algorithm_n2v),
            train_data=simple_array,
        )

    minimum_algorithm_n2v["data_type"] = SupportedData.CUSTOM.value
    with pytest.raises(ValueError):
        PredictDataModule(
            data_config=InferenceConfig(**minimum_algorithm_n2v),
            train_data=simple_array,
        )

    minimum_algorithm_n2v["data_type"] = SupportedData.ARRAY.value
    with pytest.raises(ValueError):
        PredictDataModule(
            data_config=InferenceConfig(**minimum_algorithm_n2v),
            train_data="path/to/data",
        )


def test_wrapper_unknown_type(simple_array):
    """Test that an error is raised if the data type is not supported."""
    with pytest.raises(ValueError):
        create_predict_datamodule(
            pred_data=simple_array,
            data_type="wrong_type",
            image_means=[0.5],
            image_stds=[0.1],
            axes="YX",
            batch_size=2,
        )


def test_wrapper_instantiated_with_tiling(simple_array):
    """Test that the data module is created correctly with an array."""
    # create data module
    data_module = create_predict_datamodule(
        pred_data=simple_array,
        data_type="array",
        image_means=[0.5],
        image_stds=[0.1],
        axes="YX",
        batch_size=2,
        tile_overlap=[2, 2],
        tile_size=[8, 8],
    )

    data_module.prepare_data()
    data_module.setup()
    assert len(list(data_module.predict_dataloader())) == 2


def test_wrapper_instantiated_without_tiling(simple_array):
    """Test that the data module is created correctly with an array."""
    # create data module
    data_module = create_predict_datamodule(
        pred_data=simple_array,
        data_type="array",
        image_means=[0.5],
        image_stds=[0.1],
        axes="YX",
        batch_size=2,
    )

    data_module.prepare_data()
    data_module.setup()
    assert len(list(data_module.predict_dataloader())) == 1



import pytest

from careamics.config import N2VAlgorithm
from careamics.config.support import (
    SupportedStructAxis,
    SupportedTransform,
)
from careamics.transforms import get_all_transforms

# TODO name is confusing, and the tests are probably not in the right place. They should
# probably be in the N2VManipulate or N2VAlgorithm test file.


def test_correct_transform_parameters(minimum_algorithm_n2v: dict):
    """Test that the transforms have the correct parameters.

    This is important to know that the transforms are not all instantiated as
    a generic transform.
    """
    model = N2VAlgorithm(**minimum_algorithm_n2v)

    # N2VManipulate
    params = model.n2v_config.model_dump()
    assert "roi_size" in params
    assert "masked_pixel_percentage" in params
    assert "strategy" in params
    assert "struct_mask_axis" in params
    assert "struct_mask_span" in params


def test_passing_incorrect_element(minimum_algorithm_n2v: dict):
    """Test that incorrect element in the list of transforms raises an error (
    e.g. passing un object rather than a string)."""
    minimum_algorithm_n2v["n2v_config"] = {
        "name": get_all_transforms()[SupportedTransform.XY_FLIP.value]()
    }
    with pytest.raises(ValueError):
        N2VAlgorithm(**minimum_algorithm_n2v)


def test_set_struct_mask(minimum_algorithm_n2v: dict):
    """Test that the struct mask can be set."""
    none = SupportedStructAxis.NONE.value
    vertical = SupportedStructAxis.VERTICAL.value
    horizontal = SupportedStructAxis.HORIZONTAL.value

    model = N2VAlgorithm(**minimum_algorithm_n2v)
    assert model.n2v_config.name == SupportedTransform.N2V_MANIPULATE.value
    assert model.n2v_config.struct_mask_axis == none
    assert model.n2v_config.struct_mask_span == 5

    model.n2v_config.struct_mask_axis = vertical
    model.n2v_config.struct_mask_span = 3
    assert model.n2v_config.struct_mask_axis == vertical
    assert model.n2v_config.struct_mask_span == 3

    model.n2v_config.struct_mask_axis = horizontal
    model.n2v_config.struct_mask_span = 7
    assert model.n2v_config.struct_mask_axis == horizontal
    assert model.n2v_config.struct_mask_span == 7

    model.n2v_config.struct_mask_axis = none
    model.n2v_config.struct_mask_span = 11
    assert model.n2v_config.struct_mask_axis == none
    assert model.n2v_config.struct_mask_span == 11


def test_set_struct_mask_wrong_value(minimum_algorithm_n2v: dict):
    """Test that passing a wrong struct mask axis raises an error."""
    model = N2VAlgorithm(**minimum_algorithm_n2v)
    with pytest.raises(ValueError):
        model.n2v_config.struct_mask_axis = "wrong_value"

    with pytest.raises(ValueError):
        model.n2v_config.struct_mask_span = 1



__all__ = [
    "FixedRandomPatchingStrategy",
    "PatchSpecs",
    "PatchingStrategy",
    "RandomPatchingStrategy",
    "SequentialPatchingStrategy",
]

from .patching_strategy_protocol import PatchingStrategy, PatchSpecs
from .random_patching import FixedRandomPatchingStrategy, RandomPatchingStrategy
from .sequential_patching import SequentialPatchingStrategy



"""Main CAREamics module."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("careamics")
except PackageNotFoundError:
    __version__ = "uninstalled"

__all__ = [
    "CAREamist",
    "Configuration",
    "algorithm_factory",
    "load_configuration",
    "save_configuration",
]

from .careamist import CAREamist
from .config import (
    Configuration,
    algorithm_factory,
    load_configuration,
    save_configuration,
)



"""Training configuration."""

from __future__ import annotations

from pprint import pformat
from typing import Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from .callback_model import CheckpointModel, EarlyStoppingModel


class TrainingConfig(BaseModel):
    """
    Parameters related to the training.

    Mandatory parameters are:
        - num_epochs: number of epochs, greater than 0.
        - batch_size: batch size, greater than 0.
        - augmentation: whether to use data augmentation or not (True or False).

    Attributes
    ----------
    num_epochs : int
        Number of epochs, greater than 0.
    """

    # Pydantic class configuration
    model_config = ConfigDict(
        validate_assignment=True,
    )

    num_epochs: int = Field(default=20, ge=1)
    """Number of epochs, greater than 0."""

    precision: Literal["64", "32", "16-mixed", "bf16-mixed"] = Field(default="32")
    """Numerical precision"""
    max_steps: int = Field(default=-1, ge=-1)
    """Maximum number of steps to train for. -1 means no limit."""
    check_val_every_n_epoch: int = Field(default=1, ge=1)
    """Validation step frequency."""
    enable_progress_bar: bool = Field(default=True)
    """Whether to enable the progress bar."""
    accumulate_grad_batches: int = Field(default=1, ge=1)
    """Number of batches to accumulate gradients over before stepping the optimizer."""
    gradient_clip_val: Optional[Union[int, float]] = None
    """The value to which to clip the gradient"""
    gradient_clip_algorithm: Literal["value", "norm"] = "norm"
    """The algorithm to use for gradient clipping (see lightning `Trainer`)."""
    logger: Optional[Literal["wandb", "tensorboard"]] = None
    """Logger to use during training. If None, no logger will be used. Available
    loggers are defined in SupportedLogger."""

    checkpoint_callback: CheckpointModel = CheckpointModel()
    """Checkpoint callback configuration, following PyTorch Lightning Checkpoint
    callback."""

    early_stopping_callback: Optional[EarlyStoppingModel] = Field(
        default=None, validate_default=True
    )
    """Early stopping callback configuration, following PyTorch Lightning Checkpoint
    callback."""

    def __str__(self) -> str:
        """Pretty string reprensenting the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    def has_logger(self) -> bool:
        """Check if the logger is defined.

        Returns
        -------
        bool
            Whether the logger is defined or not.
        """
        return self.logger is not None

    @field_validator("max_steps")
    @classmethod
    def validate_max_steps(cls, max_steps: int) -> int:
        """Validate the max_steps parameter.

        Parameters
        ----------
        max_steps : int
            Maximum number of steps to train for. -1 means no limit.

        Returns
        -------
        int
            Validated max_steps.
        """
        if max_steps == 0:
            raise ValueError("max_steps must be greater than 0. Use -1 for no limit.")
        return max_steps



"""CARE algorithm configuration."""

from typing import Annotated, Literal

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import AfterValidator

from careamics.config.architectures import UNetModel
from careamics.config.validators import (
    model_without_final_activation,
    model_without_n2v2,
)

from .unet_algorithm_model import UNetBasedAlgorithm

CARE = "CARE"

CARE_DESCRIPTION = (
    "Content-aware image restoration (CARE) is a deep-learning-based "
    "algorithm that uses a U-Net architecture to restore images. CARE "
    "is a supervised algorithm that requires pairs of noisy and "
    "clean images to train the network. The algorithm learns to "
    "predict the clean image from the noisy image. CARE is "
    "particularly useful for denoising images acquired in low-light "
    "conditions, such as fluorescence microscopy images."
)
CARE_REF = CiteEntry(
    text='Weigert, Martin, et al. "Content-aware image restoration: pushing the '
    'limits of fluorescence microscopy." Nature methods 15.12 (2018): 1090-1097.',
    doi="10.1038/s41592-018-0216-7",
)


class CAREAlgorithm(UNetBasedAlgorithm):
    """CARE algorithm configuration.

    Attributes
    ----------
    algorithm : "care"
        CARE Algorithm name.
    loss : {"mae", "mse"}
        CARE-compatible loss function.
    """

    algorithm: Literal["care"] = "care"
    """CARE Algorithm name."""

    loss: Literal["mae", "mse"] = "mae"
    """CARE-compatible loss function."""

    model: Annotated[
        UNetModel,
        AfterValidator(model_without_n2v2),
        AfterValidator(model_without_final_activation),
    ]
    """UNet without a final activation function and without the `n2v2` modifications."""

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the algorithm friendly name.

        Returns
        -------
        str
            Friendly name of the algorithm.
        """
        return CARE

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        return [
            "restoration",
            "UNet",
            "3D" if self.model.is_3D() else "2D",
            "CAREamics",
            "pytorch",
            CARE,
        ]

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        return CARE_REF.text + " doi: " + CARE_REF.doi

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        return [CARE_REF]

    def get_algorithm_description(self) -> str:
        """
        Get the algorithm description.

        Returns
        -------
        str
            Algorithm description.
        """
        return CARE_DESCRIPTION



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Union

from numpy.typing import DTypeLike, NDArray
from typing_extensions import Self

from careamics.dataset.dataset_utils import reshape_array
from careamics.file_io.read import ReadFunc, read_tiff


class InMemoryImageStack:
    """
    A class for extracting patches from an image stack that has been loaded into memory.
    """

    def __init__(self, source: Union[Path, Literal["array"]], data: NDArray):
        self.source: Union[Path, Literal["array"]] = source
        # data expected to be in SC(Z)YX shape, reason to use from_array constructor
        self._data: NDArray = data
        self.data_shape: Sequence[int] = self._data.shape
        self.data_dtype: DTypeLike = self._data.dtype

    def extract_patch(
        self, sample_idx: int, coords: Sequence[int], patch_size: Sequence[int]
    ) -> NDArray:
        if len(coords) != len(patch_size):
            raise ValueError("Length of coords and extent must match.")
        # TODO: test for 2D or 3D?
        return self._data[
            (
                sample_idx,  # type: ignore
                ...,  # type: ignore
                *[slice(c, c + e) for c, e in zip(coords, patch_size)],  # type: ignore
            )
        ]

    @classmethod
    def from_array(cls, data: NDArray, axes: str) -> Self:
        data = reshape_array(data, axes)
        return cls(source="array", data=data)

    @classmethod
    def from_tiff(cls, path: Path, axes: str) -> Self:
        data = read_tiff(path)
        data = reshape_array(data, axes)
        return cls(source=path, data=data)

    @classmethod
    def from_custom_file_type(
        cls, path: Path, axes: str, read_func: ReadFunc, **read_kwargs: Any
    ) -> Self:
        data = read_func(path, **read_kwargs)
        data = reshape_array(data, axes)
        return cls(source=path, data=data)



"""A module for random patching strategies."""

from collections.abc import Sequence
from typing import Optional

import numpy as np

from .patching_strategy_protocol import PatchSpecs


class RandomPatchingStrategy:
    """
    A patching strategy for sampling random patches, it implements the
    `PatchingStrategy` `Protocol`.

    The output of `get_patch_spec` will be random, i.e. if the same index is given
    twice the two outputs can be different.

    However the strategy still ensures that there will be a known number of patches for
    each sample in each image stack. This is achieved through defining a set of bins
    that map to each sample in each image stack. Whichever bin an `index` passed to
    `get_patch_spec` falls into, determines the `"data_idx"` and `"sample_idx"` in
    the returned `PatchSpecs`, but the `"coords"` will be random.

    The number of patches in each sample is based on the number of patches that would
    fit if they were sampled sequentially, non-overlapping, and covering the entire
    array.
    """

    def __init__(
        self,
        data_shapes: Sequence[Sequence[int]],
        patch_size: Sequence[int],
        seed: Optional[int] = None,
    ):
        """
        A patching strategy for sampling random patches.

        Parameters
        ----------
        data_shapes : sequence of (sequence of int)
            The shapes of the underlying data. Each element is the dimension of the
            axes SC(Z)YX.
        patch_size : sequence of int
            The size of the patch. The sequence will have length 2 or 3, for 2D and 3D
            data respectively.
        seed : int, optional
            An optional seed to ensure the reproducibility of the random patches.
        """
        self.rng = np.random.default_rng(seed=seed)
        self.patch_size = patch_size
        self.data_shapes = data_shapes

        # these bins will determine which image stack and sample a patch comes from
        # the image_stack_cumulative_patches map a patch index to each image stack
        # the sample_cumulative_patches map a patch index to each sample
        # the image_stack_cumulative_samples map a sample index to each image stack
        (
            self.image_stack_cumulative_patches,
            self.sample_cumulative_patches,
            self.image_stack_cumulative_samples,
        ) = self._calc_bins(self.data_shapes, self.patch_size)

    @property
    def n_patches(self) -> int:
        """
        The number of patches that this patching strategy will return.

        It also determines the maximum index that can be given to `get_patch_spec`.
        """
        # last bin boundary will be total patches
        return self.image_stack_cumulative_patches[-1]

    def get_patch_spec(self, index: int) -> PatchSpecs:
        """Return the patch specs for a given index.

        Parameters
        ----------
        index : int
            A patch index.

        Returns
        -------
        PatchSpecs
            A dictionary that specifies a single patch in a series of `ImageStacks`.
        """
        # TODO: break into smaller testable functions?
        if index >= self.n_patches:
            raise IndexError(
                f"Index {index} out of bounds for RandomPatchingStrategy with number "
                f"of patches {self.n_patches}"
            )
        # digitize returns the bin that `index` belongs to
        data_index = np.digitize(index, bins=self.image_stack_cumulative_patches).item()
        # maps to a particular sample within the whole series of image stacks
        #   (not just a single image stack)
        total_samples_index = np.digitize(
            index, bins=self.sample_cumulative_patches
        ).item()

        data_shape = self.data_shapes[data_index]
        spatial_shape = data_shape[2:]

        # calculate sample index relative to image stack:
        #   subtract the total number of samples in the previous image stacks
        if data_index == 0:
            n_previous_samples = 0
        else:
            n_previous_samples = self.image_stack_cumulative_samples[data_index - 1]
        sample_index = total_samples_index - n_previous_samples
        coords = _generate_random_coords(spatial_shape, self.patch_size, self.rng)
        return {
            "data_idx": data_index,
            "sample_idx": sample_index,
            "coords": coords,
            "patch_size": self.patch_size,
        }

    @staticmethod
    def _calc_bins(
        data_shapes: Sequence[Sequence[int]], patch_size: Sequence[int]
    ) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
        """Calculate bins used to map an index to an image_stack and a sample.

        The number of patches in each sample is based on the number of patches that
        would fit if they were sampled sequentially.

        Parameters
        ----------
        data_shapes : sequence of (sequence of int)
            The shapes of the underlying data. Each element is the dimension of the
            axes SC(Z)YX.
        patch_size : sequence of int
            The size of the patch. The sequence will have length 2 or 3, for 2D and 3D
            data respectively.

        Returns
        -------
        image_stack_cumulative_patches: tuple of int
            The bins that map a patch index to an image stack. E.g. if a patch index
            falls below the first bin boundary it belongs to the first image stack, if
            a patch index falls between the first bin boundary and the second bin
            boundary it belongs to the second image stack, and so on.
        sample_cumulative_patches: tuple of int
            The bins that map a patch index to a sample. E.g. if a patch index
            falls below the first bin boundary it belongs to the first sample, if
            a patch index falls between the first bin boundary and the second bin
            boundary it belongs to the second sample, and so on.
        image_stack_cumulative_samples: tuple of int
            The bins that map a sample index to an image stack. E.g. if a sample index
            falls below the first bin boundary it belongs to the first image stack, if
            a patch index falls between the first bin boundary and the second bin
            boundary it belongs to the second image stack, and so on.
        """
        patches_per_image_stack: list[int] = []
        patches_per_sample: list[int] = []
        samples_per_image_stack: list[int] = []
        for data_shape in data_shapes:
            spatial_shape = data_shape[2:]
            n_single_sample_patches = _calc_n_patches(spatial_shape, patch_size)
            # multiply by number of samples in image_stack
            patches_per_image_stack.append(n_single_sample_patches * data_shape[0])
            # list of length `sample` filled with `n_single_sample_patches`
            patches_per_sample.extend([n_single_sample_patches] * data_shape[0])
            # number of samples in each image stack
            samples_per_image_stack.append(data_shape[0])

        # cumulative sum creates the bins
        image_stack_cumulative_patches = np.cumsum(patches_per_image_stack)
        sample_cumulative_patches = np.cumsum(patches_per_sample)
        image_stack_cumulative_samples = np.cumsum(samples_per_image_stack)
        return (
            tuple(image_stack_cumulative_patches),
            tuple(sample_cumulative_patches),
            tuple(image_stack_cumulative_samples),
        )


class FixedRandomPatchingStrategy:
    """
    A patching strategy for sampling random patches it implements the `PatchingStrategy`
    `Protocol`.

    The output of `get_patch_spec` will be deterministic, i.e. if the same index is
    given twice the two outputs will be the same.

    The number of patches in each sample is based on the number of patches that would
    fit if they were sampled sequentially, non-overlapping, and covering the entire
    array.
    """

    def __init__(
        self,
        data_shapes: Sequence[Sequence[int]],
        patch_size: Sequence[int],
        seed: Optional[int] = None,
    ):
        """A patching strategy for sampling random patches.

        Parameters
        ----------
        data_shapes : sequence of (sequence of int)
            The shapes of the underlying data. Each element is the dimension of the
            axes SC(Z)YX.
        patch_size : sequence of int
            The size of the patch. The sequence will have length 2 or 3, for 2D and 3D
            data respectively.
        seed : int, optional
            An optional seed to ensure the reproducibility of the random patches.
        """
        self.rng = np.random.default_rng(seed=seed)
        self.patch_size = patch_size
        self.data_shapes = data_shapes

        # simply generate all the patches at initialisation, so they will be fixed
        self.fixed_patch_specs: list[PatchSpecs] = []
        for data_idx, data_shape in enumerate(self.data_shapes):
            spatial_shape = data_shape[2:]
            n_patches = _calc_n_patches(spatial_shape, self.patch_size)
            for sample_idx in range(data_shape[0]):
                for _ in range(n_patches):
                    random_coords = _generate_random_coords(
                        spatial_shape, self.patch_size, self.rng
                    )
                    patch_specs: PatchSpecs = {
                        "data_idx": data_idx,
                        "sample_idx": sample_idx,
                        "coords": random_coords,
                        "patch_size": self.patch_size,
                    }
                    self.fixed_patch_specs.append(patch_specs)

    @property
    def n_patches(self):
        """
        The number of patches that this patching strategy will return.

        It also determines the maximum index that can be given to `get_patch_spec`.
        """
        return len(self.fixed_patch_specs)

    def get_patch_spec(self, index: int) -> PatchSpecs:
        """Return the patch specs for a given index.

        Parameters
        ----------
        index : int
            A patch index.

        Returns
        -------
        PatchSpecs
            A dictionary that specifies a single patch in a series of `ImageStacks`.
        """
        if index >= self.n_patches:
            raise IndexError(
                f"Index {index} out of bounds for FixedRandomPatchingStrategy with "
                f"number of patches, {self.n_patches}"
            )
        # simply index the pre-generated patches to get the correct patch
        return self.fixed_patch_specs[index]


def _generate_random_coords(
    spatial_shape: Sequence[int], patch_size: Sequence[int], rng: np.random.Generator
) -> tuple[int, ...]:
    """Generate random patch coordinates for a given `spatial_shape` and `patch_size`.

    The coords are the top-left (and first z-slice for 3D data) of a patch. The
    sequence will have length 2 or 3, for 2D and 3D data respectively.

    Parameters
    ----------
    spatial_shape : sequence of int
        The dimension of the axes (Z)YX, a sequence of length 2 or 3, for 2D and 3D
        data respectively.
    patch_size : sequence of int
        The size of the patch. The sequence will have length 2 or 3, for 2D and 3D
        data respectively.
    rng : numpy.random.Generator
        A numpy generator to ensure the reproducibility of the random patches.

    Returns
    -------
    coords: tuple of int
        The top-left (and first z-slice for 3D data) coords of a patch. The tuple will
        have length 2 or 3, for 2D and 3D data respectively.

    Raises
    ------
    ValueError
        Raises if the number of spatial dimensions do not match the number of patch
        dimensions.
    """
    if len(patch_size) != len(spatial_shape):
        raise ValueError(
            f"Number of patch dimension {len(patch_size)}, do not match the number of "
            f"spatial dimensions {len(spatial_shape)}, for `patch_size={patch_size}` "
            f"and `spatial_shape={spatial_shape}`."
        )
    return tuple(
        rng.integers(
            np.zeros(len(patch_size), dtype=int),
            np.array(spatial_shape) - np.array(patch_size),
            endpoint=False,
            dtype=int,
        ).tolist()
    )


def _calc_n_patches(spatial_shape: Sequence[int], patch_size: Sequence[int]) -> int:
    """
    Calculates the number of patches for a given `spatial_shape` and `patch_size`.

    This is based on the number of patches that would fit if they were sampled
    sequentially.

    Parameters
    ----------
    spatial_shape : sequence of int
        The dimension of the axes (Z)YX, a sequence of length 2 or 3, for 2D and 3D
        data respectively.
    patch_size : sequence of int
        The size of the patch. The sequence will have length 2 or 3, for 2D and 3D
        data respectively.

    Returns
    -------
    int
        The number of patches.
    """
    if len(patch_size) != len(spatial_shape):
        raise ValueError(
            f"Number of patch dimension {len(patch_size)}, do not match the number of "
            f"spatial dimensions {len(spatial_shape)}, for `patch_size={patch_size}` "
            f"and `spatial_shape={spatial_shape}`."
        )
    return int(np.ceil(np.prod(spatial_shape) / np.prod(patch_size)))



from pathlib import Path

import pytest

from careamics.config import (
    Configuration,
    load_configuration,
    save_configuration,
)


def test_config_to_yaml(tmp_path: Path, minimum_supervised_configuration: dict):
    """Test that we can export a config to yaml and load it back"""

    # test that we can instantiate a config
    myconf = Configuration(**minimum_supervised_configuration)

    # export to yaml
    yaml_path = save_configuration(myconf, tmp_path)
    assert yaml_path.exists()

    # load from yaml
    my_other_conf = load_configuration(yaml_path)
    assert my_other_conf == myconf


def test_config_to_yaml_wrong_path(
    tmp_path: Path, minimum_supervised_configuration: dict
):
    """Test that an error is raised when the path is not a directory and not a .yml"""

    # test that we can instantiate a config
    myconf = Configuration(**minimum_supervised_configuration)

    # export to yaml
    yaml_path = tmp_path / "tmp.txt"
    with pytest.raises(ValueError):
        save_configuration(myconf, yaml_path)

    # existing file
    yaml_path.touch()
    with pytest.raises(ValueError):
        save_configuration(myconf, yaml_path)



import numpy as np
import pytest

from careamics.config import InferenceConfig
from careamics.dataset import InMemoryPredDataset


@pytest.mark.parametrize(
    "shape, axes, expected_shape",
    [
        ((16, 16), "YX", (1, 16, 16)),
        ((3, 16, 16), "CYX", (3, 16, 16)),
        ((8, 16, 16), "ZYX", (1, 8, 16, 16)),
        ((3, 8, 16, 16), "CZYX", (3, 8, 16, 16)),
        ((4, 16, 16), "SYX", (1, 16, 16)),
        ((4, 3, 16, 16), "SCYX", (3, 16, 16)),
        ((4, 3, 8, 16, 16), "SCZYX", (3, 8, 16, 16)),
    ],
)
def test_correct_normalized_outputs(shape, axes, expected_shape):
    """Test that the dataset returns normalized images with singleton
    sample dimension."""
    rng = np.random.default_rng(42)

    # check expected length
    if "S" in axes:
        # find index of S and check shape
        idx = axes.index("S")
        n_patches = shape[idx]
    else:
        n_patches = 1

    # check number of channels
    if "C" in axes:
        # get index
        idx = axes.index("C")
        n_channels = shape[idx]
    else:
        n_channels = 1

    # create array
    array = 255 * rng.random(shape)

    # create config
    config = InferenceConfig(
        data_type="array",
        axes=axes,
        image_means=[np.mean(array)] * n_channels,
        image_stds=[np.std(array)] * n_channels,
    )

    # create dataset
    dataset = InMemoryPredDataset(config, array)

    # check length
    assert len(dataset) == n_patches

    # check that the dataset returns normalized images
    for i in range(len(dataset)):
        img = dataset[i][0]

        # check that it has the correct shape
        assert img.shape == expected_shape

        # check that the image is normalized
        assert np.isclose(np.mean(img), 0, atol=0.1)
        assert np.isclose(np.std(img), 1, atol=0.1)

        # check that they are independent slices
        for j in range(i + 1, len(dataset)):
            img2 = dataset[j]
            assert not np.allclose(img, img2)



"""A script for serializers in the careamics package."""

import ast
import json
from typing import Union

import numpy as np
import torch


def _array_to_json(arr: Union[np.ndarray, torch.Tensor]) -> str:
    """Convert an array to a list and then to a JSON string.

    Parameters
    ----------
    arr : Union[np.ndarray, torch.Tensor]
        Array to be serialized.

    Returns
    -------
    str
        JSON string representing the array.
    """
    if isinstance(arr, str):
        return arr
    return json.dumps(arr.tolist())


def _to_numpy(lst: Union[str, list]) -> np.ndarray:
    """Deserialize a list or string representing a list into `np.ndarray`.

    Parameters
    ----------
    lst : list
        List or string representing a list with the array content to be deserialized.

    Returns
    -------
    np.ndarray
        The deserialized array.
    """
    if isinstance(lst, str):
        lst = ast.literal_eval(lst)
    return np.asarray(lst)


def _to_torch(lst: Union[str, list]) -> torch.Tensor:
    """Deserialize list or string representing a list into `torch.Tensor`.

    Parameters
    ----------
    lst : Union[str, list]
        List or string representing a list swith the array content to be deserialized.

    Returns
    -------
    torch.Tensor
        The deserialized tensor.
    """
    if isinstance(lst, str):
        lst = ast.literal_eval(lst)
    return torch.tensor(lst)



from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Callable, Literal, Optional

import numpy as np
import pytest
import torch
from pydantic import ValidationError
from pytorch_lightning import Trainer
from torch.utils.data import DataLoader, Dataset

from careamics.config import VAEBasedAlgorithm
from careamics.config.architectures import LVAEModel
from careamics.config.likelihood_model import (
    GaussianLikelihoodConfig,
    NMLikelihoodConfig,
)
from careamics.config.loss_model import LVAELossConfig
from careamics.config.nm_model import GaussianMixtureNMConfig, MultiChannelNMConfig
from careamics.lightning import VAEModule
from careamics.losses import denoisplit_loss, denoisplit_musplit_loss, musplit_loss
from careamics.models.lvae.likelihoods import GaussianLikelihood, NoiseModelLikelihood
from careamics.models.lvae.noise_models import (
    MultiChannelNoiseModel,
)
from careamics.utils.metrics import RunningPSNR

pytestmark = pytest.mark.lvae


# TODO: move to conftest.py as pytest.fixture
def create_dummy_noise_model(
    tmp_path: Path,
    n_gaussians: int = 3,
    n_coeffs: int = 3,
) -> None:
    weights = np.random.rand(3 * n_gaussians, n_coeffs)
    nm_dict = {
        "trained_weight": weights,
        "min_signal": np.array([0]),
        "max_signal": np.array([2**16 - 1]),
        "min_sigma": 0.125,
    }
    np.savez(tmp_path / "dummy_noise_model.npz", **nm_dict)


# TODO: move to conftest.py as pytest.fixture
# it can be split into modules for more clarity (?)
def create_split_lightning_model(
    tmp_path: Path,
    algorithm: str,
    loss_type: str,
    multiscale_count: int = 1,
    predict_logvar: Optional[Literal["pixelwise"]] = None,
    target_ch: int = 1,
) -> VAEModule:
    """Instantiate the muSplit lightining model."""
    lvae_config = LVAEModel(
        architecture="LVAE",
        input_shape=(64, 64),
        multiscale_count=multiscale_count,
        z_dims=[128, 128, 128, 128],
        output_channels=target_ch,
        predict_logvar=predict_logvar,
    )

    loss_config = LVAELossConfig(loss_type=loss_type)

    # gaussian likelihood
    if loss_type in ["musplit", "denoisplit_musplit"]:
        gaussian_lik_config = GaussianLikelihoodConfig(
            predict_logvar=predict_logvar,
            logvar_lowerbound=0.0,
        )
    else:
        gaussian_lik_config = None
    # noise model likelihood
    if loss_type in ["denoisplit", "denoisplit_musplit"]:
        create_dummy_noise_model(tmp_path, 3, 3)
        gmm = GaussianMixtureNMConfig(
            model_type="GaussianMixtureNoiseModel",
            path=tmp_path / "dummy_noise_model.npz",
        )
        noise_model_config = MultiChannelNMConfig(noise_models=[gmm] * target_ch)
        nm_lik_config = NMLikelihoodConfig()
    else:
        noise_model_config = None
        nm_lik_config = None

    vae_config = VAEBasedAlgorithm(
        algorithm=algorithm,
        loss=loss_config,
        model=lvae_config,
        gaussian_likelihood=gaussian_lik_config,
        noise_model=noise_model_config,
        noise_model_likelihood=nm_lik_config,
    )

    return VAEModule(
        algorithm_config=vae_config,
    )


class DummyDataset(Dataset):
    def __init__(
        self,
        img_size: int = 64,
        target_ch: int = 1,
        multiscale_count: int = 1,
    ):
        self.num_samples = 3
        self.img_size = img_size
        self.target_ch = target_ch
        self.multiscale_count = multiscale_count

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx: int):
        input_ = torch.randn(self.multiscale_count, self.img_size, self.img_size)
        target = torch.randn(self.target_ch, self.img_size, self.img_size)
        return input_, target


def create_dummy_dloader(
    batch_size: int = 1,
    img_size: int = 64,
    target_ch: int = 1,
    multiscale_count: int = 1,
):
    dataset = DummyDataset(
        img_size=img_size,
        target_ch=target_ch,
        multiscale_count=multiscale_count,
    )
    return DataLoader(dataset, batch_size=batch_size, shuffle=False)


@pytest.mark.parametrize(
    "multiscale_count, predict_logvar, loss_type, exp_error",
    [
        (1, None, "musplit", does_not_raise()),
        (1, "pixelwise", "musplit", does_not_raise()),
        (5, None, "musplit", does_not_raise()),
        (5, "pixelwise", "musplit", does_not_raise()),
        (1, None, "denoisplit", pytest.raises(ValueError)),
    ],
)
def test_musplit_lightining_init(
    multiscale_count: int,
    predict_logvar: str,
    loss_type: str,
    exp_error: Callable,
):
    lvae_config = LVAEModel(
        architecture="LVAE",
        input_shape=(64, 64),
        multiscale_count=multiscale_count,
        z_dims=[128, 128, 128, 128],
        output_channels=3,
        predict_logvar=predict_logvar,
    )

    likelihood_config = GaussianLikelihoodConfig(
        predict_logvar=predict_logvar,
        logvar_lowerbound=0.0,
    )

    with exp_error:
        vae_config = VAEBasedAlgorithm(
            algorithm="musplit",
            loss=LVAELossConfig(loss_type=loss_type),
            model=lvae_config,
            gaussian_likelihood=likelihood_config,
        )
        lightning_model = VAEModule(
            algorithm_config=vae_config,
        )
        assert lightning_model is not None
        assert isinstance(lightning_model.model, torch.nn.Module)
        assert lightning_model.noise_model is None
        assert lightning_model.noise_model_likelihood is None
        assert isinstance(lightning_model.gaussian_likelihood, GaussianLikelihood)
        assert lightning_model.loss_func == musplit_loss


@pytest.mark.parametrize(
    "multiscale_count, predict_logvar, target_ch, nm_cnt, loss_type, exp_error",
    [
        (1, None, 1, 1, "denoisplit", does_not_raise()),
        (5, None, 1, 1, "denoisplit", does_not_raise()),
        (1, None, 3, 3, "denoisplit", does_not_raise()),
        (5, None, 3, 3, "denoisplit", does_not_raise()),
        (1, None, 2, 4, "denoisplit", pytest.raises(ValidationError)),
        (1, None, 1, 1, "denoisplit_musplit", does_not_raise()),
        (1, None, 3, 3, "denoisplit_musplit", does_not_raise()),
        (1, "pixelwise", 1, 1, "denoisplit_musplit", does_not_raise()),
        (1, "pixelwise", 3, 3, "denoisplit_musplit", does_not_raise()),
        (5, None, 1, 1, "denoisplit_musplit", does_not_raise()),
        (5, None, 3, 3, "denoisplit_musplit", does_not_raise()),
        (5, "pixelwise", 1, 1, "denoisplit_musplit", does_not_raise()),
        (5, "pixelwise", 3, 3, "denoisplit_musplit", does_not_raise()),
        (5, "pixelwise", 2, 4, "denoisplit_musplit", pytest.raises(ValidationError)),
        (1, "pixelwise", 1, 1, "denoisplit", pytest.raises(ValidationError)),
        (1, None, 1, 1, "musplit", pytest.raises(ValueError)),
    ],
)
def test_denoisplit_lightining_init(
    tmp_path: Path,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
    nm_cnt: int,
    loss_type: str,
    exp_error: Callable,
):
    # Create the model config
    lvae_config = LVAEModel(
        architecture="LVAE",
        input_shape=(64, 64),
        multiscale_count=multiscale_count,
        z_dims=[128, 128, 128, 128],
        output_channels=target_ch,
        predict_logvar=predict_logvar,
    )

    # Create the likelihood config(s)
    # gaussian
    if loss_type == "denoisplit_musplit":
        gaussian_lik_config = GaussianLikelihoodConfig(
            predict_logvar=predict_logvar,
            logvar_lowerbound=0.0,
        )
    else:
        gaussian_lik_config = None
    # noise model
    create_dummy_noise_model(tmp_path, 3, 3)
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
    )
    noise_model_config = MultiChannelNMConfig(noise_models=[gmm] * nm_cnt)
    nm_lik_config = NMLikelihoodConfig()

    with exp_error:
        vae_config = VAEBasedAlgorithm(
            algorithm="denoisplit",
            loss=LVAELossConfig(loss_type=loss_type),
            model=lvae_config,
            gaussian_likelihood=gaussian_lik_config,
            noise_model=noise_model_config,
            noise_model_likelihood=nm_lik_config,
        )
        lightning_model = VAEModule(
            algorithm_config=vae_config,
        )
        assert lightning_model is not None
        assert isinstance(lightning_model.model, torch.nn.Module)
        assert isinstance(lightning_model.noise_model, MultiChannelNoiseModel)
        assert isinstance(lightning_model.noise_model_likelihood, NoiseModelLikelihood)
        if loss_type == "denoisplit_musplit":
            assert isinstance(lightning_model.gaussian_likelihood, GaussianLikelihood)
            assert lightning_model.loss_func == denoisplit_musplit_loss
        else:
            assert lightning_model.gaussian_likelihood is None
            assert lightning_model.loss_func == denoisplit_loss


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("multiscale_count", [1, 5])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("target_ch", [1, 3])
def test_musplit_training_step(
    batch_size: int,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
):
    lightning_model = create_split_lightning_model(
        tmp_path=None,
        algorithm="musplit",
        loss_type="musplit",
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=batch_size,
        img_size=64,
        multiscale_count=multiscale_count,
        target_ch=target_ch,
    )
    batch = next(iter(dloader))
    train_loss = lightning_model.training_step(batch=batch, batch_idx=0)

    # check outputs
    assert train_loss is not None
    assert isinstance(train_loss, dict)
    assert "loss" in train_loss
    assert "reconstruction_loss" in train_loss
    assert "kl_loss" in train_loss


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("multiscale_count", [1, 5])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("target_ch", [1, 3])
def test_musplit_validation_step(
    batch_size: int,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
):
    lightning_model = create_split_lightning_model(
        tmp_path=None,
        algorithm="musplit",
        loss_type="musplit",
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=batch_size,
        img_size=64,
        multiscale_count=multiscale_count,
        target_ch=target_ch,
    )
    batch = next(iter(dloader))
    lightning_model.validation_step(batch=batch, batch_idx=0)
    # NOTE: `validation_step` does not return anything...


@pytest.mark.parametrize(
    "multiscale_count, predict_logvar, target_ch, loss_type",
    [
        (1, None, 1, "denoisplit"),
        (5, None, 1, "denoisplit"),
        (1, None, 3, "denoisplit"),
        (5, None, 3, "denoisplit"),
        (1, None, 1, "denoisplit_musplit"),
        (1, None, 3, "denoisplit_musplit"),
        (1, "pixelwise", 1, "denoisplit_musplit"),
        (1, "pixelwise", 3, "denoisplit_musplit"),
        (5, None, 1, "denoisplit_musplit"),
        (5, None, 3, "denoisplit_musplit"),
        (5, "pixelwise", 1, "denoisplit_musplit"),
        (5, "pixelwise", 3, "denoisplit_musplit"),
    ],
)
def test_denoisplit_training_step(
    tmp_path: Path,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
    loss_type: str,
):
    lightning_model = create_split_lightning_model(
        tmp_path=tmp_path,
        algorithm="denoisplit",
        loss_type=loss_type,
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=8,
        img_size=64,
        multiscale_count=multiscale_count,
        target_ch=target_ch,
    )
    batch = next(iter(dloader))
    train_loss = lightning_model.training_step(batch=batch, batch_idx=0)

    # check outputs
    assert train_loss is not None
    assert isinstance(train_loss, dict)
    assert "loss" in train_loss
    assert "reconstruction_loss" in train_loss
    assert "kl_loss" in train_loss


@pytest.mark.parametrize(
    "multiscale_count, predict_logvar, target_ch, loss_type",
    [
        (1, None, 1, "denoisplit"),
        (5, None, 1, "denoisplit"),
        (1, None, 3, "denoisplit"),
        (5, None, 3, "denoisplit"),
        (1, None, 1, "denoisplit_musplit"),
        (1, None, 3, "denoisplit_musplit"),
        (1, "pixelwise", 1, "denoisplit_musplit"),
        (1, "pixelwise", 3, "denoisplit_musplit"),
        (5, None, 1, "denoisplit_musplit"),
        (5, None, 3, "denoisplit_musplit"),
        (5, "pixelwise", 1, "denoisplit_musplit"),
        (5, "pixelwise", 3, "denoisplit_musplit"),
    ],
)
def test_denoisplit_validation_step(
    tmp_path: Path,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
    loss_type: str,
):
    lightning_model = create_split_lightning_model(
        tmp_path=tmp_path,
        algorithm="denoisplit",
        loss_type=loss_type,
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=8,
        img_size=64,
        multiscale_count=multiscale_count,
        target_ch=target_ch,
    )
    batch = next(iter(dloader))
    lightning_model.validation_step(batch=batch, batch_idx=0)


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("multiscale_count", [1, 5])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("target_ch", [1, 3])
def test_training_loop_musplit(
    batch_size: int,
    multiscale_count: int,
    predict_logvar: str,
    target_ch: int,
):
    lightning_model = create_split_lightning_model(
        tmp_path=None,
        algorithm="musplit",
        loss_type="musplit",
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=batch_size,
        img_size=64,
        multiscale_count=multiscale_count,
        target_ch=target_ch,
    )
    trainer = Trainer(accelerator="cpu", max_epochs=2, logger=False, callbacks=[])

    try:
        trainer.fit(
            model=lightning_model,
            train_dataloaders=dloader,
            val_dataloaders=dloader,
        )
    except Exception as e:
        pytest.fail(f"Training routine failed with exception: {e}")


@pytest.mark.parametrize(
    "batch_size, predict_logvar, target_ch, loss_type",
    [
        (1, None, 1, "denoisplit"),
        (4, None, 1, "denoisplit"),
        (1, None, 3, "denoisplit"),
        (4, None, 3, "denoisplit"),
        (1, None, 1, "denoisplit_musplit"),
        (1, None, 3, "denoisplit_musplit"),
        (1, "pixelwise", 1, "denoisplit_musplit"),
        (1, "pixelwise", 3, "denoisplit_musplit"),
        (4, None, 1, "denoisplit_musplit"),
        (4, None, 3, "denoisplit_musplit"),
        (4, "pixelwise", 1, "denoisplit_musplit"),
        (4, "pixelwise", 3, "denoisplit_musplit"),
    ],
)
def test_training_loop_denoisplit(
    tmp_path: Path,
    batch_size: int,
    predict_logvar: str,
    target_ch: int,
    loss_type: str,
):
    lightning_model = create_split_lightning_model(
        tmp_path=tmp_path,
        algorithm="denoisplit",
        loss_type=loss_type,
        multiscale_count=1,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=batch_size,
        img_size=64,
        multiscale_count=1,
        target_ch=target_ch,
    )
    trainer = Trainer(accelerator="cpu", max_epochs=2, logger=False, callbacks=[])

    try:
        trainer.fit(
            model=lightning_model,
            train_dataloaders=dloader,
            val_dataloaders=dloader,
        )
    except Exception as e:
        pytest.fail(f"Training routine failed with exception: {e}")


@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("target_ch", [1, 3])
def test_get_reconstructed_tensor(
    predict_logvar: str,
    target_ch: int,
):
    lightning_model = create_split_lightning_model(
        tmp_path=None,
        algorithm="musplit",
        loss_type="musplit",
        multiscale_count=1,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    dloader = create_dummy_dloader(
        batch_size=1,
        img_size=64,
        multiscale_count=1,
        target_ch=target_ch,
    )
    input_, target = next(iter(dloader))
    output = lightning_model(input_)
    rec_img = lightning_model.get_reconstructed_tensor(output)
    assert rec_img.shape == target.shape  # same shape as target


@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("target_ch", [1, 3])
def test_val_PSNR_computation(
    predict_logvar: str,
    target_ch: int,
):
    lightning_model = create_split_lightning_model(
        tmp_path=None,
        algorithm="musplit",
        loss_type="musplit",
        multiscale_count=1,
        predict_logvar=predict_logvar,
        target_ch=target_ch,
    )
    assert lightning_model.running_psnr is not None
    assert len(lightning_model.running_psnr) == target_ch
    for item in lightning_model.running_psnr:
        assert isinstance(item, RunningPSNR)

    dloader = create_dummy_dloader(
        batch_size=1,
        img_size=64,
        multiscale_count=1,
        target_ch=target_ch,
    )
    input_, target = next(iter(dloader))
    output = lightning_model(input_)

    curr_psnr = lightning_model.compute_val_psnr(output, target)
    assert curr_psnr is not None
    assert len(curr_psnr) == target_ch
    for i in range(target_ch):
        assert lightning_model.running_psnr[i].mse_sum != 0
        assert lightning_model.running_psnr[i].N == 1
        assert lightning_model.running_psnr[i].min is not None
        assert lightning_model.running_psnr[i].max is not None
        assert lightning_model.running_psnr[i].get() is not None
        lightning_model.running_psnr[i].reset()
        assert lightning_model.running_psnr[i].mse_sum == 0
        assert lightning_model.running_psnr[i].N == 0
        assert lightning_model.running_psnr[i].min is None
        assert lightning_model.running_psnr[i].max is None
        assert lightning_model.running_psnr[i].get() is None



"""Tests for the pytest fixtures."""

from careamics.config import Configuration
from careamics.config.algorithms import UNetBasedAlgorithm
from careamics.config.data import DataConfig
from careamics.config.inference_model import InferenceConfig
from careamics.config.training_model import TrainingConfig


def test_minimum_algorithm(minimum_algorithm_n2v):
    # create algorithm configuration
    UNetBasedAlgorithm(**minimum_algorithm_n2v)


def test_minimum_data(minimum_data):
    # create data configuration
    DataConfig(**minimum_data)


def test_minimum_prediction(minimum_inference):
    # create prediction configuration
    InferenceConfig(**minimum_inference)


def test_minimum_training(minimum_training):
    # create training configuration
    TrainingConfig(**minimum_training)


def test_minimum_n2v_configuration(minimum_n2v_configuration):
    # create configuration
    Configuration(**minimum_n2v_configuration)


def test_minimum_configuration(minimum_supervised_configuration):
    # create configuration
    Configuration(**minimum_supervised_configuration)
